from django.apps import apps
from django.db.models.signals import post_migrate, post_save, pre_delete
from django.utils.translation import gettext_lazy as _

from mayan.apps.acls.classes import ModelPermission
from mayan.apps.common.apps import MayanAppConfig
from mayan.apps.common.classes import ModelCopy
from mayan.apps.common.menus import (
    menu_list_facet, menu_main, menu_multi_item, menu_object, menu_related,
    menu_return, menu_secondary, menu_setup, menu_tools
)
from mayan.apps.databases.classes import (
    ModelField, ModelProperty, ModelReverseField
)
from mayan.apps.documents.links.document_type_links import link_document_type_list
from mayan.apps.documents.signals import signal_post_document_type_change
from mayan.apps.events.classes import EventModelRegistry, ModelEventType
from mayan.apps.logging.classes import ErrorLog
from mayan.apps.logging.permissions import permission_error_log_entry_view
from mayan.apps.navigation.classes import SourceColumn
from mayan.apps.rest_api.fields import DynamicSerializerField
from mayan.apps.views.column_widgets import TwoStateWidget

from .classes import DocumentStateHelper, WorkflowAction
from .column_widgets import WorkflowLogExtraDataWidget
from .events import (
    event_workflow_instance_created, event_workflow_instance_transitioned,
    event_workflow_template_edited
)
from .handlers import (
    handler_create_workflow_image_cache, handler_launch_workflow_on_create,
    handler_launch_workflow_on_type_change, handler_trigger_transition,
    handler_workflow_template_post_edit,
    handler_workflow_template_state_post_edit,
    handler_workflow_template_state_pre_delete,
    handler_workflow_template_transition_post_edit,
    handler_workflow_template_transition_pre_delete
)
from .html_widgets import widget_transition_events
from .links import (
    link_document_multiple_workflow_templates_launch,
    link_document_single_workflow_templates_launch,
    link_document_type_workflow_template_list, link_tool_launch_workflows,
    link_workflow_instance_detail, link_workflow_instance_list,
    link_workflow_instance_transition,
    link_workflow_runtime_proxy_document_list,
    link_workflow_runtime_proxy_list,
    link_workflow_runtime_proxy_state_document_list,
    link_workflow_runtime_proxy_state_list, link_workflow_template_create,
    link_workflow_template_document_type_list, link_workflow_template_edit,
    link_workflow_template_delete_multiple, link_workflow_template_launch,
    link_workflow_template_list, link_workflow_template_delete_single,
    link_workflow_template_preview, link_workflow_template_setup,
    link_workflow_template_state_list,
    link_workflow_template_state_action_delete,
    link_workflow_template_state_action_edit,
    link_workflow_template_state_action_list,
    link_workflow_template_state_action_selection,
    link_workflow_template_state_create,
    link_workflow_template_state_delete, link_workflow_template_state_edit,
    link_workflow_template_state_escalation_create,
    link_workflow_template_state_escalation_delete,
    link_workflow_template_state_escalation_edit,
    link_workflow_template_state_escalation_list,
    link_workflow_template_transition_create,
    link_workflow_template_transition_delete,
    link_workflow_template_transition_edit,
    link_workflow_template_transition_list,
    link_workflow_template_transition_triggers,
    link_workflow_template_transition_field_create,
    link_workflow_template_transition_field_delete,
    link_workflow_template_transition_field_edit,
    link_workflow_template_transition_field_list
)
from .methods import (
    method_document_type_workflow_templates_add,
    method_document_type_workflow_templates_remove
)
from .permissions import (
    permission_workflow_template_delete, permission_workflow_template_edit,
    permission_workflow_tools, permission_workflow_instance_transition,
    permission_workflow_template_view
)


class DocumentStatesApp(MayanAppConfig):
    app_namespace = 'document_states'
    app_url = 'workflows'
    has_rest_api = True
    has_tests = True
    name = 'mayan.apps.document_states'
    verbose_name = _(message='Workflows')

    def ready(self):
        super().ready()

        Action = apps.get_model(
            app_label='actstream', model_name='Action'
        )
        Document = apps.get_model(
            app_label='documents', model_name='Document'
        )
        DocumentType = apps.get_model(
            app_label='documents', model_name='DocumentType'
        )

        Workflow = self.get_model('Workflow')
        WorkflowInstance = self.get_model('WorkflowInstance')
        WorkflowInstanceLogEntry = self.get_model('WorkflowInstanceLogEntry')
        WorkflowRuntimeProxy = self.get_model('WorkflowRuntimeProxy')
        WorkflowState = self.get_model('WorkflowState')
        WorkflowStateAction = self.get_model('WorkflowStateAction')
        WorkflowStateEscalation = self.get_model('WorkflowStateEscalation')
        WorkflowStateRuntimeProxy = self.get_model('WorkflowStateRuntimeProxy')
        WorkflowTransition = self.get_model('WorkflowTransition')
        WorkflowTransitionField = self.get_model('WorkflowTransitionField')
        WorkflowTransitionTriggerEvent = self.get_model(
            'WorkflowTransitionTriggerEvent'
        )

        Document.add_to_class(
            name='workflow', value=DocumentStateHelper.constructor
        )

        DocumentType.add_to_class(
            name='workflow_templates_add',
            value=method_document_type_workflow_templates_add
        )
        DocumentType.add_to_class(
            name='workflow_templates_remove',
            value=method_document_type_workflow_templates_remove
        )

        DynamicSerializerField.add_serializer(
            klass=Workflow,
            serializer_class='mayan.apps.document_states.serializers.workflow_template_serializers.WorkflowTemplateSerializer'
        )

        error_log = ErrorLog(app_config=self)
        error_log.register_model(model=WorkflowStateAction)

        EventModelRegistry.register(model=WorkflowInstance)
        EventModelRegistry.register(
            exclude=(WorkflowRuntimeProxy,), model=Workflow,
        )
        EventModelRegistry.register(
            exclude=(WorkflowStateRuntimeProxy,), model=WorkflowState
        )
        EventModelRegistry.register(model=WorkflowStateAction)
        EventModelRegistry.register(model=WorkflowStateEscalation)
        EventModelRegistry.register(model=WorkflowTransition)
        EventModelRegistry.register(model=WorkflowTransitionField)
        EventModelRegistry.register(model=WorkflowTransitionTriggerEvent)

        WorkflowAction.load_modules()

        ModelCopy(model=WorkflowState).add_fields(
            field_names=(
                'actions', 'workflow', 'label', 'initial', 'completion'
            )
        )
        ModelCopy(model=WorkflowStateAction).add_fields(
            field_names=(
                'state', 'label', 'enabled', 'when', 'backend_path',
                'backend_data', 'condition'
            )
        )
        ModelCopy(model=WorkflowTransition).add_fields(
            field_names=(
                'workflow', 'label', 'origin_state', 'destination_state',
                'condition', 'fields', 'trigger_events'
            ),
            field_value_gets={
                'origin_state': {
                    'workflow': '{workflow.pk}',
                    'label': '{instance.origin_state.label}'
                },
                'destination_state': {
                    'workflow': '{workflow.pk}',
                    'label': '{instance.destination_state.label}'
                }
            }
        )
        ModelCopy(
            model=WorkflowTransitionTriggerEvent
        ).add_fields(
            field_names=('transition', 'event_type')
        )
        ModelCopy(
            model=WorkflowTransitionField
        ).add_fields(
            field_names=(
                'transition', 'field_type', 'name', 'label', 'help_text',
                'required', 'widget', 'widget_kwargs',
            )
        )
        ModelCopy(
            model=Workflow, bind_link=True, register_permission=True
        ).add_fields(
            field_names=(
                'auto_launch', 'internal_name', 'label', 'document_types',
                'states', 'transitions'
            )
        )

        ModelEventType.register(
            event_types=(
                event_workflow_instance_created,
                event_workflow_instance_transitioned,
            ), model=Document
        )
        ModelEventType.register(
            event_types=(event_workflow_template_edited,), model=Workflow
        )
        ModelEventType.register(
            event_types=(
                event_workflow_instance_created,
                event_workflow_instance_transitioned,
            ), model=WorkflowInstance
        )

        ModelProperty(
            model=Document,
            name='workflow.< workflow internal name >.get_current_state',
            label=_(message='Current state of a workflow'), description=_(
                'Return the current state of the selected workflow.'
            )
        )
        ModelProperty(
            model=Document,
            name='workflow.< workflow internal name >.get_current_state.completion',
            label=_(message='Current state of a workflow'), description=_(
                'Return the completion value of the current state of the '
                'selected workflow.'
            )
        )

        ModelPermission.register(
            model=Document, permissions=(
                permission_workflow_instance_transition,
                permission_workflow_template_view,
                permission_workflow_tools
            )
        )
        ModelPermission.register(
            model=Workflow, permissions=(
                permission_error_log_entry_view, permission_workflow_template_delete,
                permission_workflow_template_edit, permission_workflow_tools,
                permission_workflow_instance_transition,
                permission_workflow_template_view
            )
        )
        ModelPermission.register(
            model=WorkflowTransition,
            permissions=(permission_workflow_instance_transition,)
        )

        ModelPermission.register_inheritance(
            model=WorkflowInstance, related='workflow',
        )
        ModelPermission.register_inheritance(
            model=WorkflowInstanceLogEntry,
            related='workflow_instance__workflow',
        )
        ModelPermission.register_inheritance(
            model=WorkflowState, related='workflow',
        )
        ModelPermission.register_inheritance(
            model=WorkflowStateAction, related='state__workflow',
        )
        ModelPermission.register_inheritance(
            model=WorkflowStateEscalation, related='state__workflow',
        )
        ModelPermission.register_inheritance(
            model=WorkflowTransition, related='workflow',
        )
        ModelPermission.register_inheritance(
            model=WorkflowTransitionField, related='transition',
        )
        ModelPermission.register_inheritance(
            model=WorkflowTransitionTriggerEvent,
            related='transition__workflow',
        )

        ModelField(model=WorkflowInstance, name='document')
        ModelField(model=WorkflowInstance, name='workflow')
        ModelReverseField(model=WorkflowInstance, name='log_entries')

        ModelProperty(
            description=_(
                'Return the last workflow instance log entry. The '
                'log entry itself has the following fields: datetime, '
                'transition, user, and comment.'
            ), label=_(message='Get last log entry'), model=WorkflowInstance,
            name='get_last_log_entry'
        )
        ModelProperty(
            description=_(
                'Return the current context dictionary which includes '
                'runtime data from the workflow transition fields.'
            ), label=_(message='Get the context'), model=WorkflowInstance,
            name='get_runtime_context'
        )
        ModelProperty(
            description=_(
                'Return the transition of the workflow instance.'
            ), label=_(message='Get last transition'), model=WorkflowInstance,
            name='get_last_transition'
        )

        SourceColumn(
            attribute='label', is_identifier=True, is_sortable=True,
            source=Workflow
        )
        column_workflow_internal_name = SourceColumn(
            attribute='internal_name', include_label=True, is_sortable=True,
            source=Workflow
        )
        column_workflow_internal_name.add_exclude(
            source=WorkflowRuntimeProxy
        )
        column_workflow_get_initial_state = SourceColumn(
            attribute='get_initial_state', empty_value=_(message='None'),
            include_label=True, source=Workflow
        )
        column_workflow_get_initial_state.add_exclude(
            source=WorkflowRuntimeProxy
        )
        SourceColumn(
            attribute='get_current_state', include_label=True,
            label=_(message='Current state'), source=WorkflowInstance,
        )
        SourceColumn(
            func=lambda context: getattr(
                context['object'].get_last_log_entry(), 'user', _(message='None')
            ), include_label=True, label=_(message='User'), source=WorkflowInstance
        )
        SourceColumn(
            attribute='get_last_transition', include_label=True,
            label=_(message='Last transition'), source=WorkflowInstance
        )
        SourceColumn(
            func=lambda context: getattr(
                context['object'].get_last_log_entry(), 'datetime', _(message='None')
            ), include_label=True, label=_(message='Date and time'),
            source=WorkflowInstance
        )
        SourceColumn(
            func=lambda context: getattr(
                context['object'].get_current_state(),
                'completion', _(message='None')
            ), include_label=True, label=_(message='Completion'),
            source=WorkflowInstance
        )

        SourceColumn(
            attribute='datetime', is_identifier=True,
            label=_(message='Date and time'), source=WorkflowInstanceLogEntry
        )
        SourceColumn(
            attribute='user', include_label=True, label=_(message='User'),
            source=WorkflowInstanceLogEntry
        )
        SourceColumn(
            attribute='transition__origin_state', include_label=True,
            is_sortable=True, source=WorkflowInstanceLogEntry
        )
        SourceColumn(
            attribute='transition', include_label=True, is_sortable=True,
            source=WorkflowInstanceLogEntry
        )
        SourceColumn(
            attribute='transition__destination_state', include_label=True,
            is_sortable=True, source=WorkflowInstanceLogEntry
        )
        SourceColumn(
            attribute='comment', include_label=True, is_sortable=True,
            source=WorkflowInstanceLogEntry
        )
        SourceColumn(
            attribute='get_extra_data', include_label=True,
            label=_(message='Additional details'), source=WorkflowInstanceLogEntry,
            widget=WorkflowLogExtraDataWidget
        )

        SourceColumn(
            attribute='label', is_identifier=True, is_sortable=True,
            source=WorkflowState
        )
        SourceColumn(
            attribute='initial', include_label=True, is_sortable=True,
            source=WorkflowState, widget=TwoStateWidget
        )
        SourceColumn(
            attribute='completion', include_label=True, is_sortable=True,
            source=WorkflowState
        )
        column_workflow_actions = SourceColumn(
            attribute='get_actions_display', include_label=True,
            source=WorkflowState
        )
        column_workflow_actions.add_exclude(
            source=WorkflowStateRuntimeProxy
        )
        column_workflow_escalations = SourceColumn(
            attribute='get_escalations_display', include_label=True,
            source=WorkflowState
        )
        column_workflow_escalations.add_exclude(
            source=WorkflowStateRuntimeProxy
        )

        SourceColumn(
            attribute='label', is_identifier=True, is_sortable=True,
            source=WorkflowStateAction
        )
        SourceColumn(
            attribute='enabled', include_label=True, is_sortable=True,
            source=WorkflowStateAction, widget=TwoStateWidget
        )
        SourceColumn(
            attribute='get_when_display', include_label=True,
            label=_(message='When?'), source=WorkflowStateAction
        )
        SourceColumn(
            attribute='get_backend_class_label', include_label=True,
            label=_(message='Action type'), source=WorkflowStateAction
        )
        SourceColumn(
            attribute='has_condition', include_label=True,
            source=WorkflowStateAction, widget=TwoStateWidget
        )

        # Workflow Template State Escalation

        SourceColumn(
            attribute='priority', is_identifier=True, is_sortable=True,
            source=WorkflowStateEscalation
        )
        SourceColumn(
            attribute='enabled', include_label=True, is_sortable=True,
            source=WorkflowStateEscalation, widget=TwoStateWidget
        )
        SourceColumn(
            attribute='unit', include_label=True,
            source=WorkflowStateEscalation
        )
        SourceColumn(
            attribute='amount', include_label=True,
            source=WorkflowStateEscalation
        )
        SourceColumn(
            attribute='transition', include_label=True, is_sortable=True,
            source=WorkflowStateEscalation
        )
        SourceColumn(
            attribute='has_condition', include_label=True,
            source=WorkflowStateEscalation, widget=TwoStateWidget
        )

        # Workflow transition

        SourceColumn(
            attribute='label', is_identifier=True, is_sortable=True,
            source=WorkflowTransition,
        )
        SourceColumn(
            attribute='origin_state', include_label=True, is_sortable=True,
            source=WorkflowTransition
        )
        SourceColumn(
            attribute='destination_state', include_label=True,
            is_sortable=True, source=WorkflowTransition
        )
        SourceColumn(
            attribute='has_condition', include_label=True,
            source=WorkflowTransition, widget=TwoStateWidget
        )
        SourceColumn(
            attribute='get_field_display', include_label=True,
            source=WorkflowTransition
        )
        SourceColumn(
            func=lambda context: widget_transition_events(
                transition=context['object']
            ), help_text=_(
                'Triggers are system events that will cause the transition '
                'to be applied.'
            ), include_label=True, label=_(message='Triggers'),
            source=WorkflowTransition
        )

        SourceColumn(
            attribute='name', is_identifier=True, is_sortable=True,
            source=WorkflowTransitionField
        )
        SourceColumn(
            attribute='label', include_label=True, is_sortable=True,
            source=WorkflowTransitionField
        )
        SourceColumn(
            attribute='get_field_type_display', include_label=True,
            label=_(message='Type'), source=WorkflowTransitionField
        )
        SourceColumn(
            attribute='required', include_label=True, is_sortable=True,
            source=WorkflowTransitionField, widget=TwoStateWidget
        )
        SourceColumn(
            attribute='get_widget_display', include_label=True,
            label=_(message='Widget'), is_sortable=False,
            source=WorkflowTransitionField
        )
        SourceColumn(
            attribute='widget_kwargs', include_label=True, is_sortable=True,
            source=WorkflowTransitionField
        )

        SourceColumn(
            func=lambda context: context['object'].get_document_count(
                user=context['request'].user
            ), include_label=True, label=_(message='Documents'), order=99,
            source=WorkflowRuntimeProxy
        )
        SourceColumn(
            func=lambda context: context['object'].get_document_count(
                user=context['request'].user
            ), include_label=True, label=_(message='Documents'), order=99,
            source=WorkflowStateRuntimeProxy
        )

        # Workflow instance

        menu_list_facet.bind_links(
            links=(link_workflow_instance_list,), sources=(Document,)
        )
        menu_object.bind_links(
            links=(
                link_workflow_instance_detail,
                link_workflow_instance_transition
            ), sources=(WorkflowInstance,)
        )
        menu_secondary.bind_links(
            links=(link_document_single_workflow_templates_launch,),
            sources=(
                WorkflowInstance,
                'document_states:document_multiple_workflow_templates_launch',
                'document_states:document_single_workflow_templates_launch',
                'document_states:workflow_instance_list'
            )
        )

        # Workflow runtime proxy

        menu_list_facet.bind_links(
            links=(
                link_workflow_runtime_proxy_document_list,
                link_workflow_runtime_proxy_state_list,
            ), sources=(WorkflowRuntimeProxy,)
        )
        menu_list_facet.bind_links(
            links=(
                link_workflow_runtime_proxy_state_document_list,
            ), sources=(WorkflowStateRuntimeProxy,)
        )

        # Workflow template

        menu_main.bind_links(
            links=(link_workflow_runtime_proxy_list,), position=20
        )
        menu_list_facet.bind_links(
            links=(
                link_workflow_template_transition_triggers,
                link_workflow_template_transition_field_list,
            ), sources=(WorkflowTransition,)
        )
        menu_list_facet.bind_links(
            links=(
                link_document_type_workflow_template_list,
            ), sources=(DocumentType,)
        )
        menu_list_facet.bind_links(
            exclude=(WorkflowRuntimeProxy,),
            links=(
                link_workflow_template_document_type_list,
                link_workflow_template_state_list,
                link_workflow_template_transition_list,
                link_workflow_template_preview
            ), sources=(Workflow,)
        )
        menu_multi_item.bind_links(
            links=(link_document_multiple_workflow_templates_launch,),
            sources=(Document,)
        )
        menu_multi_item.bind_links(
            links=(link_workflow_template_delete_multiple,),
            sources=(Workflow,)
        )
        menu_object.bind_links(
            exclude=(WorkflowRuntimeProxy,), links=(
                link_workflow_template_delete_single,
                link_workflow_template_edit,
                link_workflow_template_launch
            ), sources=(Workflow,)
        )
        menu_related.bind_links(
            links=(link_workflow_template_list,), sources=(
                DocumentType, 'documents:document_type_list',
                'documents:document_type_create'
            )
        )
        menu_related.bind_links(
            exclude=(WorkflowRuntimeProxy,),
            links=(link_document_type_list,),
            sources=(
                Workflow, 'document_states:workflow_template_create',
                'document_states:workflow_template_list'
            )
        )
        menu_return.bind_links(
            exclude=(WorkflowRuntimeProxy,), links=(
                link_workflow_template_list,
            ), sources=(
                Workflow, 'document_states:workflow_template_create',
                'document_states:workflow_template_list'
            )
        )
        menu_secondary.bind_links(
            exclude=(WorkflowRuntimeProxy,), links=(
                link_workflow_template_create,
            ), sources=(
                Workflow, 'document_states:workflow_template_create',
                'document_states:workflow_template_list'
            )
        )
        menu_setup.bind_links(
            links=(link_workflow_template_setup,)
        )
        menu_tools.bind_links(
            links=(link_tool_launch_workflows,)
        )

        # Workflow template escalation

        menu_object.bind_links(
            links=(
                link_workflow_template_state_escalation_edit,
                link_workflow_template_state_escalation_delete,
            ), sources=(WorkflowStateEscalation,)
        )

        # Workflow template state

        menu_list_facet.bind_links(
            exclude=(WorkflowStateRuntimeProxy,),
            links=(
                link_workflow_template_state_action_list,
                link_workflow_template_state_escalation_list
            ), sources=(WorkflowState,)
        )
        menu_object.bind_links(
            exclude=(WorkflowStateRuntimeProxy,),
            links=(
                link_workflow_template_state_edit,
                link_workflow_template_state_delete
            ), sources=(WorkflowState,)
        )
        menu_secondary.bind_links(
            links=(
                link_workflow_template_state_action_selection,
                link_workflow_template_state_escalation_create
            ), sources=(WorkflowState,)
        )
        menu_secondary.bind_links(
            links=(link_workflow_template_state_create,), sources=(
                WorkflowState,
                'document_states:workflow_template_state_create',
                'document_states:workflow_template_state_list'
            )
        )

        # Workflow template state action

        menu_object.bind_links(
            links=(
                link_workflow_template_state_action_edit,
                link_workflow_template_state_action_delete,
            ), sources=(WorkflowStateAction,)
        )

        # Workflow template transition

        menu_object.bind_links(
            links=(
                link_workflow_template_transition_edit,
                link_workflow_template_transition_delete
            ), sources=(WorkflowTransition,)
        )
        menu_secondary.bind_links(
            links=(link_workflow_template_transition_field_create,),
            sources=(WorkflowTransition,)
        )

        # Workflow template transition field

        menu_object.bind_links(
            links=(
                link_workflow_template_transition_field_delete,
                link_workflow_template_transition_field_edit
            ), sources=(WorkflowTransitionField,)
        )
        menu_secondary.bind_links(
            links=(link_workflow_template_transition_create,), sources=(
                WorkflowTransition,
                'document_states:workflow_template_transition_create',
                'document_states:workflow_template_transition_list'
            )
        )

        post_migrate.connect(
            dispatch_uid='workflows_handler_create_workflow_image_cache',
            receiver=handler_create_workflow_image_cache
        )
        post_save.connect(
            dispatch_uid='workflows_handler_launch_workflow_on_create',
            receiver=handler_launch_workflow_on_create,
            sender=Document
        )
        signal_post_document_type_change.connect(
            dispatch_uid='workflows_handler_launch_workflow_on_type_change',
            receiver=handler_launch_workflow_on_type_change,
            sender=Document
        )

        # Indexing, general

        post_save.connect(
            dispatch_uid='workflows_handler_trigger_transition',
            receiver=handler_trigger_transition,
            sender=Action
        )

        # Indexing, Workflow template

        post_save.connect(
            dispatch_uid='workflows_handler_workflow_template_post_edit',
            receiver=handler_workflow_template_post_edit,
            sender=Workflow
        )

        # Indexing, Workflow template state

        post_save.connect(
            dispatch_uid='workflows_handler_workflow_template_state_post_edit',
            receiver=handler_workflow_template_state_post_edit,
            sender=WorkflowState
        )
        pre_delete.connect(
            dispatch_uid='workflows_handler_workflow_template_state_pre_delete',
            receiver=handler_workflow_template_state_pre_delete,
            sender=WorkflowState
        )

        # Indexing, Workflow template transition

        post_save.connect(
            dispatch_uid='workflows_handler_workflow_template_transition_post_edit',
            receiver=handler_workflow_template_transition_post_edit,
            sender=WorkflowTransition
        )
        pre_delete.connect(
            dispatch_uid='workflows_handler_workflow_template_transition_pre_delete',
            receiver=handler_workflow_template_transition_pre_delete,
            sender=WorkflowTransition
        )



from .utils import get_context_user


def condition_is_current_user(context, resolved_object):
    if condition_user_is_authenticated(context=context, resolved_object=resolved_object):
        user = get_context_user(context=context)

        if user:
            return resolved_object == user


def condition_not_is_current_user(context, resolved_object):
    return condition_user_is_authenticated(
        context=context, resolved_object=resolved_object
    ) and not condition_is_current_user(
        context=context, resolved_object=resolved_object
    )


def condition_user_is_authenticated(context, resolved_object):
    user = get_context_user(context=context)

    if user:
        return user.is_authenticated



import logging

from django.urls import resolve as django_resolve
from django.urls.base import get_script_prefix

logger = logging.getLogger(name=__name__)


def convert_to_id_list(items):
    return ','.join(
        map(str, items)
    )


def request_is_ajax(request):
    return request.META.get('HTTP_X_REQUESTED_WITH') == 'XMLHttpRequest'


def resolve(path, urlconf=None):
    path = '/{}'.format(
        path.replace(
            get_script_prefix(), '', 1
        )
    )
    return django_resolve(path=path, urlconf=urlconf)



from django.urls import re_path

from .api_views import (
    APICabinetDocumentAddView, APICabinetDocumentListView,
    APICabinetDocumentRemoveView, APICabinetListView, APICabinetView,
    APIDocumentCabinetListView
)
from .views import (
    CabinetChildAddView, CabinetCreateView, CabinetDeleteView,
    CabinetDetailView, CabinetEditView, CabinetListView,
    DocumentCabinetAddView, DocumentCabinetListView,
    DocumentCabinetRemoveView
)

urlpatterns_cabinets = [
    re_path(
        route=r'^cabinets/$', name='cabinet_list',
        view=CabinetListView.as_view()
    ),
    re_path(
        route=r'^cabinets/create/$', name='cabinet_create',
        view=CabinetCreateView.as_view()
    ),
    re_path(
        route=r'^cabinets/(?P<cabinet_id>\d+)/$', name='cabinet_view',
        view=CabinetDetailView.as_view()
    ),
    re_path(
        route=r'^cabinets/(?P<cabinet_id>\d+)/children/add/$',
        name='cabinet_child_add', view=CabinetChildAddView.as_view()
    ),
    re_path(
        route=r'^cabinets/(?P<cabinet_id>\d+)/delete/$',
        name='cabinet_delete', view=CabinetDeleteView.as_view()
    ),
    re_path(
        route=r'^cabinets/(?P<cabinet_id>\d+)/edit/$', name='cabinet_edit',
        view=CabinetEditView.as_view()
    )
]

urlpatterns_documents_cabinets = [
    re_path(
        route=r'^documents/(?P<document_id>\d+)/cabinets/$',
        name='document_cabinet_list', view=DocumentCabinetListView.as_view()
    ),
    re_path(
        route=r'^documents/(?P<document_id>\d+)/cabinets/add/$',
        name='document_cabinet_add', view=DocumentCabinetAddView.as_view()
    ),
    re_path(
        route=r'^documents/multiple/cabinets/add/$',
        name='document_multiple_cabinet_add',
        view=DocumentCabinetAddView.as_view()
    ),
    re_path(
        route=r'^documents/(?P<document_id>\d+)/cabinets/remove/$',
        name='document_cabinet_remove',
        view=DocumentCabinetRemoveView.as_view()
    ),
    re_path(
        route=r'^documents/multiple/cabinets/remove/$',
        name='multiple_document_cabinet_remove',
        view=DocumentCabinetRemoveView.as_view()
    )
]

urlpatterns = []
urlpatterns.extend(urlpatterns_cabinets)
urlpatterns.extend(urlpatterns_documents_cabinets)

api_urls = [
    re_path(
        route=r'^cabinets/(?P<cabinet_id>[0-9]+)/$', name='cabinet-detail',
        view=APICabinetView.as_view()
    ),
    re_path(
        route=r'^cabinets/$', name='cabinet-list',
        view=APICabinetListView.as_view()
    ),
    re_path(
        route=r'^cabinets/(?P<cabinet_id>[0-9]+)/documents/$',
        name='cabinet-document-list',
        view=APICabinetDocumentListView.as_view()
    ),
    re_path(
        route=r'^cabinets/(?P<cabinet_id>[0-9]+)/documents/add/$',
        name='cabinet-document-add', view=APICabinetDocumentAddView.as_view()
    ),
    re_path(
        route=r'^cabinets/(?P<cabinet_id>[0-9]+)/documents/remove/$',
        name='cabinet-document-remove', view=APICabinetDocumentRemoveView.as_view()
    ),
    re_path(
        route=r'^documents/(?P<document_id>[0-9]+)/cabinets/$',
        name='document-cabinet-list',
        view=APIDocumentCabinetListView.as_view()
    )
]



from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('contenttypes', '0002_remove_content_type_name')
    ]

    operations = [
        migrations.CreateModel(
            name='Message',
            fields=[
                (
                    'id', models.AutoField(
                        auto_created=True, primary_key=True, serialize=False,
                        verbose_name='ID'
                    )
                ),
                (
                    'sender_object_id', models.PositiveIntegerField(
                        blank=True, null=True
                    )
                ),
                (
                    'subject', models.CharField(
                        help_text='Short description of this message.',
                        max_length=255, verbose_name='Subject'
                    )
                ),
                (
                    'body', models.TextField(
                        help_text='The actual content of the message.',
                        verbose_name='Body'
                    )
                ),
                (
                    'read', models.BooleanField(
                        default=False, help_text='This field determines '
                        'if the message has been read or not.',
                        verbose_name='Read'
                    )
                ),
                (
                    'date_time', models.DateTimeField(
                        auto_now_add=True, help_text='Date and time of '
                        'the message creation.',
                        verbose_name='Creation date and time'
                    )
                ),
                (
                    'parent', models.ForeignKey(
                        blank=True, null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='messages', to='messaging.Message',
                        verbose_name='Parent message'
                    )
                ),
                (
                    'sender_content_type', models.ForeignKey(
                        blank=True, null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to='contenttypes.ContentType'
                    )
                ),
                (
                    'user', models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='messages', to=settings.AUTH_USER_MODEL,
                        verbose_name='User'
                    )
                )
            ],
            options={
                'verbose_name': 'Message',
                'verbose_name_plural': 'Messages',
                'ordering': ('-date_time',)
            }
        )
    ]



from ..literals import FIELD_TYPE_CHOICE_CHAR, WORKFLOW_ACTION_ON_ENTRY

DOCUMENT_WORKFLOW_LAUNCH_ACTION_CLASS_PATH = 'mayan.apps.document_states.workflow_actions.DocumentWorkflowLaunchAction'

TEST_DOCUMENT_EDIT_WORKFLOW_TEMPLATE_STATE_ACTION_DOTTED_PATH = 'mayan.apps.document_states.workflow_actions.DocumentPropertiesEditAction'
TEST_DOCUMENT_EDIT_WORKFLOW_TEMPLATE_STATE_ACTION_TEXT_LABEL = 'new document label'
TEST_DOCUMENT_EDIT_WORKFLOW_TEMPLATE_STATE_ACTION_TEXT_DESCRIPTION = 'new document description'
TEST_DOCUMENT_EDIT_WORKFLOW_TEMPLATE_STATE_ACTION_TEXT_DATA = {
    'document_label': TEST_DOCUMENT_EDIT_WORKFLOW_TEMPLATE_STATE_ACTION_TEXT_LABEL,
    'document_description': TEST_DOCUMENT_EDIT_WORKFLOW_TEMPLATE_STATE_ACTION_TEXT_DESCRIPTION
}
TEST_DOCUMENT_EDIT_WORKFLOW_TEMPLATE_STATE_ACTION_TEMPLATE_LABEL = '{{ workflow_instance.document.label }} new'
TEST_DOCUMENT_EDIT_WORKFLOW_TEMPLATE_STATE_ACTION_TEMPLATE_DESCRIPTION = '{{ workflow_instance.document.label }}'
TEST_DOCUMENT_EDIT_WORKFLOW_TEMPLATE_STATE_ACTION_TEMPLATE_DATA = {
    'document_label': TEST_DOCUMENT_EDIT_WORKFLOW_TEMPLATE_STATE_ACTION_TEMPLATE_LABEL,
    'document_description': TEST_DOCUMENT_EDIT_WORKFLOW_TEMPLATE_STATE_ACTION_TEMPLATE_DESCRIPTION
}

TEST_HEADERS_KEY = 'test key'
TEST_HEADERS_VALUE = 'test value'
TEST_HEADERS_JSON = '{{"{}": "{}"}}'.format(
    TEST_HEADERS_KEY, TEST_HEADERS_VALUE
)
TEST_HEADERS_JSON_TEMPLATE_KEY = 'test key'
TEST_HEADERS_JSON_TEMPLATE_VALUE = '{{ workflow_instance.document.label }}'
TEST_HEADERS_JSON_TEMPLATE = '{{"{}": "{}"}}'.format(
    TEST_HEADERS_JSON_TEMPLATE_KEY, TEST_HEADERS_JSON_TEMPLATE_VALUE
)
TEST_HEADERS_AUTHENTICATION_KEY = 'Authorization'
TEST_HEADERS_AUTHENTICATION_VALUE = 'Basic dGVzdHVzZXJuYW1lOnRlc3RwYXNzd29yZA=='
TEST_PAYLOAD_JSON = '{"label": "label"}'
TEST_PAYLOAD_TEMPLATE_DOCUMENT_LABEL = '{"label": "{{ workflow_instance.document.label }}"}'
TEST_SERVER_USERNAME = 'testusername'
TEST_SERVER_PASSWORD = 'testpassword'

TEST_WORKFLOW_INSTANCE_LOG_ENTRY_COMMENT = 'test workflow instance log entry comment'
TEST_WORKFLOW_INSTANCE_LOG_ENTRY_EXTRA_DATA = '{"test": "test"}'

TEST_WORKFLOW_TEMPLATE_INTERNAL_NAME = 'test_workflow_template_name'
TEST_WORKFLOW_TEMPLATE_LABEL = 'test workflow template label'
TEST_WORKFLOW_TEMPLATE_LABEL_EDITED = 'test workflow template label edited'
TEST_WORKFLOW_TEMPLATE_INITIAL_STATE_LABEL = 'test initial state'
TEST_WORKFLOW_TEMPLATE_INITIAL_STATE_COMPLETION = 33

TEST_WORKFLOW_INDEX_TEMPLATE_EXPRESSION = '{% if  document.workflows.first %}{{ document.workflows.first.workflow.label }}-{{ document.workflows.first.get_current_state }}{% endif %}'
TEST_WORKFLOW_TEMPLATE_TRANSITION_INDEX_TEMPLATE_EXPRESSION = '{% if  document.workflows.first %}{{ document.workflows.first.workflow.label }}-{{ document.workflows.first.get_current_state }}-{{ document.workflows.first.workflow.transitions.first }}{% endif %}'

TEST_WORKFLOW_TEMPLATE_STATE_ACTION_LABEL = 'test template state action label'
TEST_WORKFLOW_TEMPLATE_STATE_ACTION_LABEL_EDITED = 'test template state action label edited'
TEST_WORKFLOW_TEMPLATE_STATE_ACTION_WHEN = WORKFLOW_ACTION_ON_ENTRY
TEST_WORKFLOW_TEMPLATE_STATE_ACTION_DOTTED_PATH = TEST_DOCUMENT_EDIT_WORKFLOW_TEMPLATE_STATE_ACTION_DOTTED_PATH
TEST_WORKFLOW_TEMPLATE_STATE_ACTION_GENERIC_DOTTED_PATH = 'mayan.apps.document_states.tests.workflow_actions.WorkflowActionTest'

TEST_WORKFLOW_TEMPLATE_STATE_LABEL = 'test workflow template state label'
TEST_WORKFLOW_TEMPLATE_STATE_LABEL_EDITED = 'test workflow template state label edited'
TEST_WORKFLOW_TEMPLATE_STATE_COMPLETION = 66

TEST_WORKFLOW_TEMPLATE_STATE_ESCALATION_COMMENT = 'test_workflow_template_state_escalation_comment'
TEST_WORKFLOW_TEMPLATE_STATE_ESCALATION_UNIT = 'minutes'
TEST_WORKFLOW_TEMPLATE_STATE_ESCALATION_AMOUNT = 1
TEST_WORKFLOW_TEMPLATE_STATE_ESCALATION_AMOUNT_EDITED = 2

TEST_WORKFLOW_TEMPLATE_TRANSITION_FIELD_HELP_TEXT = 'test workflow template transition field help test'
TEST_WORKFLOW_TEMPLATE_TRANSITION_FIELD_LABEL = 'test workflow template transition field'
TEST_WORKFLOW_TEMPLATE_TRANSITION_FIELD_LABEL_EDITED = 'test workflow template transition field edited'
TEST_WORKFLOW_TEMPLATE_TRANSITION_FIELD_NAME = 'test_workflow_template_transition_field'
TEST_WORKFLOW_TEMPLATE_TRANSITION_FIELD_TYPE = FIELD_TYPE_CHOICE_CHAR

TEST_WORKFLOW_TEMPLATE_TRANSITION_LABEL = 'test workflow template transition label'
TEST_WORKFLOW_TEMPLATE_TRANSITION_LABEL_EDITED = 'test workflow template transition label edited'



from datetime import timedelta
import re

from dateutil.parser import parse

from django.template import Library, Node, TemplateSyntaxError
from django.utils.html import strip_spaces_between_tags

from ..utils import process_regex_flags

register = Library()


@register.filter
def date_parse(date_string):
    """
    Takes a string and converts it into a datetime object.
    """
    return parse(timestr=date_string)


@register.filter
def dict_get(dictionary, key):
    """
    Return the value for the given key or '' if not found.
    """
    return dictionary.get(key, '')


@register.simple_tag
def method(obj, method, *args, **kwargs):
    """
    Call an object method. {% method object method **kwargs %}
    """
    try:
        return getattr(obj, method)(*args, **kwargs)
    except Exception as exception:
        raise TemplateSyntaxError(
            'Error calling object method; {}'.format(exception)
        )


@register.simple_tag
def regex_findall(pattern, string, **kwargs):
    """
    Return all non-overlapping matches of pattern in string, as a list of
    strings. {% regex_findall pattern string flags %}
    """
    flags = process_regex_flags(**kwargs)
    return re.findall(flags=flags, pattern=pattern, string=string)


@register.simple_tag
def regex_match(pattern, string, **kwargs):
    """
    If zero or more characters at the beginning of string match the regular
    expression pattern, return a corresponding match object.
    {% regex_match pattern string flags %}
    """
    flags = process_regex_flags(**kwargs)
    return re.match(flags=flags, pattern=pattern, string=string)


@register.simple_tag
def regex_search(pattern, string, **kwargs):
    """
    Scan through string looking for the first location where the regular
    expression pattern produces a match, and return a corresponding
    match object. {% regex_search pattern string flags %}
    """
    flags = process_regex_flags(**kwargs)
    return re.search(flags=flags, pattern=pattern, string=string)


@register.simple_tag
def regex_sub(pattern, repl, string, count=0, **kwargs):
    """
    Replacing the leftmost non-overlapping occurrences of pattern in
    string with repl. {% regex_sub pattern repl string count=0 flags %}
    """
    flags = process_regex_flags(**kwargs)
    return re.sub(
        count=count, flags=flags, pattern=pattern, repl=repl, string=string
    )


@register.simple_tag
def set(value):
    """
    Set a context variable to a specific value.
    """
    return value


@register.filter
def split(obj, separator):
    """
    Return a list of the words in the string, using sep as the delimiter
    string.
    """
    return obj.split(separator)


class SpacelessPlusNode(Node):
    def __init__(self, nodelist):
        self.nodelist = nodelist

    def render(self, context):
        content = self.nodelist.render(context).strip()
        result = []
        for line in content.split('\n'):
            if line.strip() != '':
                result.append(line)

        return strip_spaces_between_tags(
            value='\n'.join(result)
        )


@register.tag
def spaceless_plus(parser, token):
    """
    Removes empty lines between the tag nodes.
    """
    nodelist = parser.parse(
        ('endspaceless_plus',)
    )
    parser.delete_first_token()
    return SpacelessPlusNode(nodelist=nodelist)


@register.simple_tag(name='timedelta')
def tag_timedelta(date, **kwargs):
    """
    Takes a datetime object and applies a timedelta.
    """
    return date + timedelta(**kwargs)



from django.core.management.base import BaseCommand

from ...models import AutoAdminSingleton


class Command(BaseCommand):
    help = 'Used to create a super user with a secure and automatic password.'

    def handle(self, *args, **options):
        AutoAdminSingleton.objects.create_autoadmin()



from django.core.files import File

from mayan.apps.sources.source_backend_actions.interfaces import SourceBackendActionInterface
from mayan.apps.storage.models import SharedUploadedFile

from ..source_backend_actions.interfaces import SourceBackendActionInterfaceTask

from .arguments import argument_file_object, argument_shared_uploaded_file_id


class SourceBackendActionMixinFileUser:
    accept_files = True

    class Interface:
        class Model(SourceBackendActionInterface):
            class Argument:
                file_object = argument_file_object

            def process_interface_context(self):
                super().process_interface_context()

                self.action_kwargs['file_object'] = File(
                    file=self.context['file_object']
                )

        class Task(SourceBackendActionInterfaceTask):
            class Argument:
                shared_uploaded_file_id = argument_shared_uploaded_file_id

            def process_interface_context(self):
                super().process_interface_context()

                self.action_kwargs['shared_uploaded_file_id'] = self.context[
                    'shared_uploaded_file_id'
                ]

    def _background_task(self, shared_uploaded_file_id, **kwargs):
        result = super()._background_task(**kwargs)

        result['server_upload_entry_list'] = (
            {'shared_uploaded_file_id': shared_uploaded_file_id},
        )

        return result

    def _execute(self, file_object, **kwargs):
        self.shared_uploaded_file = SharedUploadedFile.objects.create(
            file=file_object
        )

        super()._execute(**kwargs)

    def get_task_kwargs(self, **kwargs):
        result = super().get_task_kwargs(**kwargs)

        result['action_interface_kwargs']['shared_uploaded_file_id'] = self.shared_uploaded_file.pk

        return result



from mayan.apps.smart_settings.classes import Setting
from mayan.apps.testing.tests.base import BaseTestCase

from ..settings import setting_storage_backend_arguments


class DocumentSignaturesSettingMigrationTestCase(BaseTestCase):
    def test_signatures_storage_backend_arguments_0001_migration(self):
        test_value = {'location': 'test value'}
        self._test_setting = setting_storage_backend_arguments
        self._test_configuration_value = '{}'.format(
            Setting.serialize_value(value=test_value)
        )
        self._create_test_configuration_file()

        self.assertEqual(
            setting_storage_backend_arguments.value, test_value
        )

    def test_signatures_storage_backend_arguments_0001_migration_with_dict(self):
        test_value = {'location': 'test value'}
        self._test_setting = setting_storage_backend_arguments
        self._test_configuration_value = test_value
        self._create_test_configuration_file()

        self.assertEqual(
            setting_storage_backend_arguments.value, test_value
        )



import copy

from django.utils.translation import gettext_lazy as _

from mayan.apps.acls.links import link_acl_list
from mayan.apps.navigation.classes import Link
from mayan.apps.navigation.utils import factory_condition_queryset_access

from .icons import (
    icon_cabinet_child_add, icon_cabinet_create, icon_cabinet_delete,
    icon_cabinet_edit, icon_cabinet_list, icon_cabinet_view,
    icon_document_cabinet_add, icon_document_cabinet_remove,
    icon_document_multiple_cabinet_add
)
from .permissions import (
    permission_cabinet_add_document, permission_cabinet_create,
    permission_cabinet_delete, permission_cabinet_edit,
    permission_cabinet_remove_document, permission_cabinet_view
)


def condition_cabinet_is_root(context, resolved_object):
    return context['resolved_object'].is_root_node()


# Document links

link_document_cabinet_list = Link(
    args='resolved_object.pk', icon=icon_cabinet_list,
    permission=permission_cabinet_view, text=_(message='Cabinets'),
    view='cabinets:document_cabinet_list'
)
link_document_cabinet_remove = Link(
    args='resolved_object.pk', icon=icon_document_cabinet_remove,
    permission=permission_cabinet_remove_document,
    text=_(message='Remove from cabinets'), view='cabinets:document_cabinet_remove'
)
link_document_cabinet_add = Link(
    args='object.pk', icon=icon_document_cabinet_add,
    permission=permission_cabinet_add_document,
    text=_(message='Add to cabinets'), view='cabinets:document_cabinet_add'
)
link_document_multiple_cabinet_add = Link(
    icon=icon_document_multiple_cabinet_add, text=_(message='Add to cabinets'),
    view='cabinets:document_multiple_cabinet_add'
)
link_multiple_document_cabinet_remove = Link(
    icon=icon_document_cabinet_remove, text=_(message='Remove from cabinets'),
    view='cabinets:multiple_document_cabinet_remove'
)

# Cabinet links

link_custom_acl_list = copy.copy(link_acl_list)
link_custom_acl_list.condition = condition_cabinet_is_root

link_cabinet_child_add = Link(
    args='object.pk', icon=icon_cabinet_child_add,
    permission=permission_cabinet_create, text=_(message='Add new level'),
    view='cabinets:cabinet_child_add'
)
link_cabinet_create = Link(
    icon=icon_cabinet_create, permission=permission_cabinet_create,
    text=_(message='Create cabinet'), view='cabinets:cabinet_create'
)
link_cabinet_delete = Link(
    args='object.pk', icon=icon_cabinet_delete,
    permission=permission_cabinet_delete, tags='dangerous',
    text=_(message='Delete'), view='cabinets:cabinet_delete'
)
link_cabinet_edit = Link(
    args='object.pk', icon=icon_cabinet_edit,
    permission=permission_cabinet_edit, text=_(message='Edit'),
    view='cabinets:cabinet_edit'
)
link_cabinet_list = Link(
    condition=factory_condition_queryset_access(
        app_label='cabinets', model_name='Cabinet',
        object_permission=permission_cabinet_view,
    ), icon=icon_cabinet_list,
    text=_(message='All'), view='cabinets:cabinet_list'
)
link_cabinet_view = Link(
    args='object.pk', icon=icon_cabinet_view,
    permission=permission_cabinet_view, text=_(message='Details'),
    view='cabinets:cabinet_view'
)



from mayan.apps.documents.events import (
    event_document_created, event_document_file_created,
    event_document_file_edited, event_document_version_created,
    event_document_version_edited, event_document_version_page_created
)
from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.permissions import permission_document_create
from mayan.apps.documents.tests.base import GenericDocumentViewTestCase
from mayan.apps.sources.tests.mixins.wizard_mixins import SourceDocumentUploadWizardTestMixin

from ..events import event_cabinet_document_added
from ..permissions import permission_cabinet_add_document

from .mixins import (
    CabinetDocumentUploadWizardStepTestMixin, CabinetTestMixin
)


class CabinetDocumentUploadViewTestCase(
    CabinetDocumentUploadWizardStepTestMixin, GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_cabinet()
        self._create_test_cabinet()

    def test_post_view_with_document_type_access_source_access(self):
        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_create
        )
        self.grant_access(
            obj=self._test_document_type,
            permission=permission_cabinet_add_document
        )
        self.grant_access(
            obj=self._test_source, permission=permission_document_create
        )

        self._clear_events()

        response = self._request_test_source_document_upload_post_view_with_cabinets()
        self.assertEqual(response.status_code, 302)

        self.assertFalse(
            self._test_cabinet_list[0] in Document.objects.first().cabinets.all()
        )
        self.assertFalse(
            self._test_cabinet_list[1] in Document.objects.first().cabinets.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 6)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active
        test_document_version_page = test_document_version.pages.first()

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_document_file)
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        self.assertEqual(events[3].action_object, test_document)
        self.assertEqual(events[3].actor, self._test_case_user)
        self.assertEqual(events[3].target, test_document_version)
        self.assertEqual(events[3].verb, event_document_version_created.id)

        self.assertEqual(events[4].action_object, test_document_version)
        self.assertEqual(events[4].actor, self._test_case_user)
        self.assertEqual(events[4].target, test_document_version_page)
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[5].action_object, test_document)
        self.assertEqual(events[5].actor, self._test_case_user)
        self.assertEqual(events[5].target, test_document_version)
        self.assertEqual(events[5].verb, event_document_version_edited.id)

    def test_post_view_with_cabinet_single_access_source_access(self):
        self.grant_access(
            obj=self._test_cabinet_list[0],
            permission=permission_cabinet_add_document
        )
        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_create
        )
        self.grant_access(
            obj=self._test_source, permission=permission_document_create
        )

        self._clear_events()

        response = self._request_test_source_document_upload_post_view_with_cabinets()
        self.assertEqual(response.status_code, 302)

        self.assertFalse(
            self._test_cabinet_list[0] in Document.objects.first().cabinets.all()
        )
        self.assertFalse(
            self._test_cabinet_list[1] in Document.objects.first().cabinets.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 6)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active
        test_document_version_page = test_document_version.pages.first()

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_document_file)
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        self.assertEqual(events[3].action_object, test_document)
        self.assertEqual(events[3].actor, self._test_case_user)
        self.assertEqual(events[3].target, test_document_version)
        self.assertEqual(events[3].verb, event_document_version_created.id)

        self.assertEqual(events[4].action_object, test_document_version)
        self.assertEqual(events[4].actor, self._test_case_user)
        self.assertEqual(events[4].target, test_document_version_page)
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[5].action_object, test_document)
        self.assertEqual(events[5].actor, self._test_case_user)
        self.assertEqual(events[5].target, test_document_version)
        self.assertEqual(events[5].verb, event_document_version_edited.id)

    def test_post_view_with_document_type_access_cabinet_single_access_source_access(self):
        self._create_test_cabinet()
        self._create_test_cabinet()

        self.grant_access(
            obj=self._test_cabinet_list[0],
            permission=permission_cabinet_add_document
        )
        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_create
        )
        self.grant_access(
            obj=self._test_document_type,
            permission=permission_cabinet_add_document
        )
        self.grant_access(
            obj=self._test_source, permission=permission_document_create
        )

        self._clear_events()

        response = self._request_test_source_document_upload_post_view_with_cabinets()
        self.assertEqual(response.status_code, 302)

        self.assertTrue(
            self._test_cabinet_list[0] in Document.objects.first().cabinets.all()
        )
        self.assertFalse(
            self._test_cabinet_list[1] in Document.objects.first().cabinets.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 7)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active
        test_document_version_page = test_document_version.pages.first()

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(
            events[1].action_object, self._test_cabinet_list[0]
        )
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, test_document)
        self.assertEqual(events[1].verb, event_cabinet_document_added.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_document_file)
        self.assertEqual(events[2].verb, event_document_file_created.id)

        self.assertEqual(events[3].action_object, test_document)
        self.assertEqual(events[3].actor, self._test_case_user)
        self.assertEqual(events[3].target, test_document_file)
        self.assertEqual(events[3].verb, event_document_file_edited.id)

        self.assertEqual(events[4].action_object, test_document)
        self.assertEqual(events[4].actor, self._test_case_user)
        self.assertEqual(events[4].target, test_document_version)
        self.assertEqual(events[4].verb, event_document_version_created.id)

        self.assertEqual(events[5].action_object, test_document_version)
        self.assertEqual(events[5].actor, self._test_case_user)
        self.assertEqual(events[5].target, test_document_version_page)
        self.assertEqual(
            events[5].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[6].action_object, test_document)
        self.assertEqual(events[6].actor, self._test_case_user)
        self.assertEqual(events[6].target, test_document_version)
        self.assertEqual(events[6].verb, event_document_version_edited.id)

    def test_post_view_with_document_type_access_cabinet_multiple_access_source_access(self):
        self._create_test_cabinet()
        self._create_test_cabinet()

        self.grant_access(
            obj=self._test_cabinet_list[0], permission=permission_cabinet_add_document
        )
        self.grant_access(
            obj=self._test_cabinet_list[1], permission=permission_cabinet_add_document
        )
        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_create
        )
        self.grant_access(
            obj=self._test_document_type, permission=permission_cabinet_add_document
        )
        self.grant_access(
            obj=self._test_source, permission=permission_document_create
        )

        self._clear_events()

        response = self._request_test_source_document_upload_post_view_with_cabinets()
        self.assertEqual(response.status_code, 302)

        self.assertTrue(
            self._test_cabinet_list[0] in Document.objects.first().cabinets.all()
        )
        self.assertTrue(
            self._test_cabinet_list[1] in Document.objects.first().cabinets.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 8)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active
        test_document_version_page = test_document_version.pages.first()

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(
            events[1].action_object, self._test_cabinet_list[0]
        )
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, test_document)
        self.assertEqual(events[1].verb, event_cabinet_document_added.id)

        self.assertEqual(
            events[2].action_object, self._test_cabinet_list[1]
        )
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_document)
        self.assertEqual(events[2].verb, event_cabinet_document_added.id)

        self.assertEqual(events[3].action_object, test_document)
        self.assertEqual(events[3].actor, self._test_case_user)
        self.assertEqual(events[3].target, test_document_file)
        self.assertEqual(events[3].verb, event_document_file_created.id)

        self.assertEqual(events[4].action_object, test_document)
        self.assertEqual(events[4].actor, self._test_case_user)
        self.assertEqual(events[4].target, test_document_file)
        self.assertEqual(events[4].verb, event_document_file_edited.id)

        self.assertEqual(events[5].action_object, test_document)
        self.assertEqual(events[5].actor, self._test_case_user)
        self.assertEqual(events[5].target, test_document_version)
        self.assertEqual(events[5].verb, event_document_version_created.id)

        self.assertEqual(events[6].action_object, test_document_version)
        self.assertEqual(events[6].actor, self._test_case_user)
        self.assertEqual(events[6].target, test_document_version_page)
        self.assertEqual(
            events[6].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[7].action_object, test_document)
        self.assertEqual(events[7].actor, self._test_case_user)
        self.assertEqual(events[7].target, test_document_version)
        self.assertEqual(events[7].verb, event_document_version_edited.id)


class CabinetStepDocumentUploadWizardTestCase(
    CabinetTestMixin, SourceDocumentUploadWizardTestMixin,
    GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_cabinet()
        self._create_test_cabinet()

    def test_post_view_with_document_type_access_source_access(self):
        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_create
        )
        self.grant_access(
            obj=self._test_document_type,
            permission=permission_cabinet_add_document
        )
        self.grant_access(
            obj=self._test_source, permission=permission_document_create
        )

        self._clear_events()

        response = self._request_document_upload_wizard_post_view()
        self.assertNotContains(
            response=response, status_code=200,
            text=self._test_cabinet_list[0].label
        )
        self.assertNotContains(
            response=response, status_code=200,
            text=self._test_cabinet_list[1].label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_post_view_with_access_cabinet_single_access_source_access(self):
        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_create
        )
        self.grant_access(
            obj=self._test_source, permission=permission_document_create
        )
        self.grant_access(
            obj=self._test_cabinet_list[0],
            permission=permission_cabinet_add_document
        )

        self._clear_events()

        response = self._request_document_upload_wizard_post_view()
        self.assertNotContains(
            response=response, status_code=200,
            text=self._test_cabinet_list[0].label
        )
        self.assertNotContains(
            response=response, status_code=200,
            text=self._test_cabinet_list[1].label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_post_view_with_document_type_access_cabinet_single_access_source_access(self):
        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_create
        )
        self.grant_access(
            obj=self._test_document_type,
            permission=permission_cabinet_add_document
        )
        self.grant_access(
            obj=self._test_source, permission=permission_document_create
        )
        self.grant_access(
            obj=self._test_cabinet_list[0],
            permission=permission_cabinet_add_document
        )

        self._clear_events()

        response = self._request_document_upload_wizard_post_view()
        self.assertContains(
            response=response, status_code=200,
            text=self._test_cabinet_list[0].label
        )
        self.assertNotContains(
            response=response, status_code=200,
            text=self._test_cabinet_list[1].label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_post_view_with_document_type_access_cabinet_multuple_access_source_access(self):
        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_create
        )
        self.grant_access(
            obj=self._test_document_type,
            permission=permission_cabinet_add_document
        )
        self.grant_access(
            obj=self._test_source, permission=permission_document_create
        )
        self.grant_access(
            obj=self._test_cabinet_list[0],
            permission=permission_cabinet_add_document
        )
        self.grant_access(
            obj=self._test_cabinet_list[1],
            permission=permission_cabinet_add_document
        )

        self._clear_events()

        response = self._request_document_upload_wizard_post_view()
        self.assertContains(
            response=response, status_code=200,
            text=self._test_cabinet_list[0].label
        )
        self.assertContains(
            response=response, status_code=200,
            text=self._test_cabinet_list[1].label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('converter', '0016_auto_20190814_0510')
    ]

    operations = [
        migrations.AlterField(
            model_name='layertransformation',
            name='name',
            field=models.CharField(max_length=128, verbose_name='Name'),
        )
    ]



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('locales', '0006_alter_userlocaleprofile_timezone')
    ]

    operations = [
        migrations.AlterField(
            model_name='userlocaleprofile', name='language',
            field=models.CharField(
                choices=[
                    ('sq', 'Albanian'),
                    ('ar', 'Arabic'),
                    ('ar-eg', 'Arabic (Egypt)'),
                    ('bs', 'Bosnian'),
                    ('bg', 'Bulgarian'),
                    ('ca', 'Catalan'),
                    ('zh-cn', 'Chinese (China)'),
                    ('zh-hans', 'Chinese (Simplified)'),
                    ('zh-tw', 'Chinese (Taiwan)'),
                    ('hr', 'Croatian'),
                    ('cs', 'Czech'),
                    ('da', 'Danish'),
                    ('nl', 'Dutch'),
                    ('en', 'English'),
                    ('fr', 'French'),
                    ('de-at', 'German (Austria)'),
                    ('de-de', 'German (Germany)'),
                    ('el', 'Greek'),
                    ('he-il', 'Hebrew (Israel)'),
                    ('hu', 'Hungarian'),
                    ('id', 'Indonesian'),
                    ('it', 'Italian'),
                    ('lv', 'Latvian'),
                    ('mn-mn', 'Mongolian (Mongolia)'),
                    ('fa', 'Persian'),
                    ('pl', 'Polish'),
                    ('pt', 'Portuguese'),
                    ('pt-br', 'Portuguese (Brazil)'),
                    ('ro-ro', 'Romanian (Romania)'),
                    ('ru', 'Russian'),
                    ('sl', 'Slovenian'),
                    ('es', 'Spanish'),
                    ('es-mx', 'Spanish (Mexico)'),
                    ('es-pr', 'Spanish (Puerto Rico)'),
                    ('th', 'Thai'),
                    ('tr', 'Turkish'),
                    ('tr-tr', 'Turkish (Turkey)'),
                    ('uk', 'Ukrainian'),
                    ('vi', 'Vietnamese')
                ], max_length=8, verbose_name='Language'
            )
        )
    ]



from django.http import HttpRequest

from mayan.apps.documents.serializers.document_serializers import (
    DocumentFileSerializer, DocumentVersionSerializer
)
from mayan.apps.documents.tests.base import GenericDocumentTestCase


class SerializerTestMixin:
    def setUp(self):
        super().setUp()
        self._test_request = HttpRequest()
        self._test_request.META['SERVER_NAME'] = '127.0.0.1'
        self._test_request.META['SERVER_PORT'] = '80'


class DocumentFileSerializerTestCase(
    SerializerTestMixin, GenericDocumentTestCase
):
    def test_document_url(self):
        serializer = DocumentFileSerializer(
            context={'request': self._test_request},
            instance=self._test_document_file
        )

        self.assertFalse(
            self._test_document.pk == self._test_document_file.pk,
            msg='Rerun test to ensure document and document file do not '
            'have the same ID.'
        )
        self.assertTrue(
            str(self._test_document.pk) in serializer.data['document_url']
        )


class DocumentVersionSerializerTestCase(
    SerializerTestMixin, GenericDocumentTestCase
):
    def test_document_url(self):
        serializer = DocumentVersionSerializer(
            context={'request': self._test_request},
            instance=self._test_document_version
        )

        self.assertFalse(
            self._test_document.pk == self._test_document_version.pk,
            msg='Rerun test to ensure document and document version do '
            'not have the same ID.'
        )
        self.assertTrue(
            str(self._test_document.pk) in serializer.data['document_url']
        )



from django.utils.translation import gettext_lazy as _

from mayan.apps.navigation.classes import Link, Separator

from .icons import icon_current_user_oidc_details
from .settings import setting_oidc_user_profile_url


def condition_no_usable_password(context, resolved_object):
    if context['request'].user.is_authenticated:
        return not (
            context[
                'request'
            ].user.has_usable_password()
        )
    else:
        return False


link_current_user_oidc_details = Link(
    condition=condition_no_usable_password, tags='new_window',
    icon=icon_current_user_oidc_details, text=_(message='OIDC user details'),
    url=setting_oidc_user_profile_url.value
)

separator_oidc_user = Separator()



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('metadata', '0007_auto_20150918_0800'),
        ('sources', '0008_auto_20150815_0351')
    ]

    operations = [
        migrations.AddField(
            model_name='emailbasemodel',
            name='from_metadata_type',
            field=models.ForeignKey(
                blank=True,
                help_text='Select a metadata type valid for the document '
                'type selected in which to store the email\'s "from" value.',
                null=True, on_delete=models.CASCADE,
                related_name='email_from', to='metadata.MetadataType',
                verbose_name='From metadata type'
            ),
            preserve_default=True
        ),
        migrations.AddField(
            model_name='emailbasemodel',
            name='subject_metadata_type',
            field=models.ForeignKey(
                blank=True, on_delete=models.CASCADE,
                related_name='email_subject',
                help_text='Select a metadata '
                'type valid for the document type selected in which to '
                'store the email\'s subject.',
                null=True, to='metadata.MetadataType',
                verbose_name='Subject metadata type'
            ),
            preserve_default=True
        )
    ]



from django.utils.translation import gettext_lazy as _

from mayan.apps.permissions.classes import PermissionNamespace

namespace = PermissionNamespace(
    label=_(message='Events'), name='events'
)

permission_events_clear = namespace.add_permission(
    label=_(message='Clear the events of an object'), name='events_clear'
)
permission_events_export = namespace.add_permission(
    label=_(message='Export the events of an object'), name='events_export'
)
permission_events_view = namespace.add_permission(
    label=_(message='View the events of an object'), name='events_view'
)



from mayan.apps.testing.tests.base import BaseTestCase

from ..exceptions import DynamicSearchModelException
from ..search_fields import (
    SearchField, SearchFieldDirect, SearchFieldRelated
)

from .mixins.base import SearchTestMixin, TestSearchObjectHierarchyTestMixin


class SearchFieldDirectDectionTestCase(
    TestSearchObjectHierarchyTestMixin, SearchTestMixin, BaseTestCase
):
    def test_detection(self):
        search_field = SearchField.init(
            field='label', search_model=self._test_search_grandparent
        )

        self.assertTrue(
            isinstance(search_field, SearchFieldDirect)
        )

    def test_detection_invalid(self):
        with self.assertRaises(expected_exception=DynamicSearchModelException):
            SearchField.init(
                field='invalid', search_model=self._test_search_grandparent
            )

    def test_get_instance_value(self):
        search_field = SearchField.init(
            field='label', search_model=self._test_search_grandparent
        )
        result = search_field.get_instance_value(
            instance=self._test_object_grandparent,
            search_backend=self._test_search_backend
        )

        self.assertEqual(result, self._test_object_grandparent.label)


class SearchFieldRelatedDectionTestCase(
    TestSearchObjectHierarchyTestMixin, SearchTestMixin, BaseTestCase
):
    def test_detection(self):
        search_field = SearchField.init(
            field='children__label',
            search_model=self._test_search_grandparent
        )

        self.assertTrue(
            isinstance(search_field, SearchFieldRelated)
        )

    def test_detection_invalid(self):
        with self.assertRaises(expected_exception=DynamicSearchModelException):
            SearchField.init(
                field='children__invalid',
                search_model=self._test_search_grandparent
            )

    def test_get_instance_value(self):
        search_field = SearchField.init(
            field='children__label',
            search_model=self._test_search_grandparent
        )
        result = search_field.get_instance_value(
            instance=self._test_object_grandparent,
            search_backend=self._test_search_backend
        )

        self.assertEqual(
            result, [self._test_object_parent.label]
        )



from django.apps import apps
from django.db.models.signals import pre_delete
from django.utils.translation import gettext_lazy as _

from mayan.apps.acls.classes import ModelPermission
from mayan.apps.acls.permissions import (
    permission_acl_edit, permission_acl_view
)
from mayan.apps.common.apps import MayanAppConfig
from mayan.apps.common.classes import MissingItem
from mayan.apps.common.menus import (
    menu_list_facet, menu_object, menu_return, menu_secondary, menu_setup
)
from mayan.apps.common.signals import signal_post_upgrade
from mayan.apps.converter.links import link_transformation_list
from mayan.apps.databases.classes import ModelProperty
from mayan.apps.documents.menus import menu_documents
from mayan.apps.documents.permissions import (
    permission_document_create, permission_document_file_new
)
from mayan.apps.events.classes import EventModelRegistry, ModelEventType
from mayan.apps.logging.classes import ErrorLog
from mayan.apps.navigation.classes import SourceColumn
from mayan.apps.rest_api.fields import DynamicSerializerField
from mayan.apps.views.column_widgets import TwoStateWidget

from .classes import DocumentCreateWizardStep
from .events import event_source_edited
from .handlers import (
    handler_delete_interval_source_periodic_task,
    handler_initialize_periodic_tasks
)
from .links import (
    link_document_file_source_metadata_list, link_document_file_upload,
    link_document_upload_wizard, link_source_backend_selection,
    link_source_delete, link_source_edit, link_source_list,
    link_source_setup, link_source_test
)
from .permissions import (
    permission_sources_delete, permission_sources_edit,
    permission_sources_metadata_view, permission_sources_view
)
from .property_helpers import (
    DocumentSourceMetadataHelper, DocumentFileSourceMetadataHelper
)
from .source_backends.base import SourceBackend


class SourcesApp(MayanAppConfig):
    app_namespace = 'sources'
    app_url = 'sources'
    has_rest_api = True
    has_static_media = False
    has_tests = True
    name = 'mayan.apps.sources'
    verbose_name = _(message='Sources')

    def ready(self):
        super().ready()

        DocumentCreateWizardStep.load_modules()

        Document = apps.get_model(
            app_label='documents', model_name='Document'
        )
        DocumentFile = apps.get_model(
            app_label='documents', model_name='DocumentFile'
        )
        DocumentType = apps.get_model(
            app_label='documents', model_name='DocumentType'
        )
        DocumentFileSourceMetadata = self.get_model(
            model_name='DocumentFileSourceMetadata'
        )
        Source = self.get_model(model_name='Source')

        SourceBackend.load_modules()

        DynamicSerializerField.add_serializer(
            klass=Source,
            serializer_class='mayan.apps.sources.serializers.SourceSerializer'
        )

        Document.add_to_class(
            name='source_metadata_value_of',
            value=DocumentSourceMetadataHelper.constructor
        )

        DocumentFile.add_to_class(
            name='source_metadata_value_of',
            value=DocumentFileSourceMetadataHelper.constructor
        )

        error_log = ErrorLog(app_config=self)
        error_log.register_model(model=Source)

        EventModelRegistry.register(model=Source)

        ModelEventType.register(
            model=Source, event_types=(event_source_edited,)
        )

        MissingItem(
            label=_(message='Create a document source'),
            description=_(
                'Document sources are the way in which new documents are '
                'feed to Mayan EDMS, create at least a web form source to '
                'be able to upload documents from a browser.'
            ),
            condition=lambda: not Source.objects.exists(),
            view='sources:source_list'
        )

        ModelPermission.register(
            model=DocumentFile, permissions=(
                permission_sources_metadata_view,
            )
        )
        ModelPermission.register(
            model=Source, permissions=(
                permission_acl_edit, permission_acl_view,
                permission_document_create, permission_document_file_new,
                permission_sources_delete, permission_sources_edit,
                permission_sources_metadata_view, permission_sources_view
            )
        )

        ModelProperty(
            description=_(
                'Return the value of a specific source metadata for '
                'the document\'s latest file.'
            ), label=_(message='Source metadata value of'), model=Document,
            name='source_metadata_value_of.< key >'
        )

        ModelProperty(
            description=_(
                'Return the value of a specific source metadata.'
            ), label=_(message='Source metadata value of'), model=DocumentFile,
            name='source_metadata_value_of.< key >'
        )

        SourceColumn(
            attribute='key', is_identifier=True,
            source=DocumentFileSourceMetadata
        )
        SourceColumn(
            attribute='value', include_label=True,
            source=DocumentFileSourceMetadata
        )
        SourceColumn(
            attribute='label', is_identifier=True, is_sortable=True,
            source=Source
        )
        SourceColumn(
            attribute='get_backend_class_label', include_label=True,
            label=_(message='Type'), source=Source
        )
        SourceColumn(
            attribute='enabled', include_label=True, is_sortable=True,
            source=Source, widget=TwoStateWidget
        )

        menu_documents.bind_links(
            links=(link_document_upload_wizard,)
        )

        menu_list_facet.bind_links(
            links=(link_document_file_source_metadata_list,),
            sources=(DocumentFile,)
        )
        menu_list_facet.bind_links(
            links=(link_transformation_list,), sources=(Source,)
        )

        menu_object.bind_links(
            links=(link_source_delete, link_source_edit), sources=(Source,)
        )

        menu_object.bind_links(
            links=(link_source_test,), sources=(Source,)
        )
        menu_return.bind_links(
            links=(link_source_list,), sources=(
                Source, 'sources:source_backend_selection',
                'sources:source_create', 'sources:source_list'
            )
        )
        menu_secondary.bind_links(
            links=(link_source_backend_selection,), sources=(
                Source, 'sources:source_backend_selection',
                'sources:source_create', 'sources:source_list'
            )
        )
        menu_secondary.bind_links(
            links=(link_document_file_upload,),
            sources=(
                'documents:document_file_list',
                'sources:document_file_upload'
            )
        )
        menu_setup.bind_links(
            links=(link_source_setup,)
        )
        pre_delete.connect(
            dispatch_uid='sources_handler_delete_interval_source_periodic_task',
            receiver=handler_delete_interval_source_periodic_task,
            sender=DocumentType
        )
        signal_post_upgrade.connect(
            dispatch_uid='sources_handler_initialize_periodic_tasks',
            receiver=handler_initialize_periodic_tasks
        )



from mayan.apps.documents.tests.mixins.document_mixins import DocumentTestMixin


class DocumentVersionExportAPIViewTestMixin(DocumentTestMixin):
    def _request_test_document_version_export_api_view_via_get(self):
        return self.get(
            viewname='rest_api:documentversion-export', kwargs={
                'document_id': self._test_document.pk,
                'document_version_id': self._test_document.version_active.pk
            }
        )

    def _request_test_document_version_export_api_view_via_post(self):
        return self.post(
            viewname='rest_api:documentversion-export', kwargs={
                'document_id': self._test_document.pk,
                'document_version_id': self._test_document.version_active.pk
            }
        )


class DocumentVersionExportViewTestMixin(DocumentTestMixin):
    def _request_test_document_version_export_view(self):
        return self.post(
            viewname='document_exports:document_version_export', kwargs={
                'document_version_id': self._test_document_version.pk
            }
        )



from django.utils.translation import gettext_lazy as _

from mayan.apps.permissions.classes import PermissionNamespace

namespace = PermissionNamespace(
    label=_(message='Search'), name='search'
)

permission_search_tools = namespace.add_permission(
    label=_(message='Execute search tools'), name='search_tools'
)



import logging

from django.apps import apps
from django.conf import settings
from django.contrib.auth import get_user_model
from django.db import OperationalError

from mayan.apps.lock_manager.backends.base import LockingBackend
from mayan.apps.lock_manager.exceptions import LockError
from mayan.celery import app

from .literals import DEFAULT_SOURCES_LOCK_EXPIRE

logger = logging.getLogger(name=__name__)


@app.task(bind=True, ignore_result=True, retry_backoff=True)
def task_source_backend_action_background_task(
    self, action_name, source_id, action_interface_kwargs=None
):
    Source = apps.get_model(
        app_label='sources', model_name='Source'
    )

    action_interface_kwargs = action_interface_kwargs or {}

    try:
        source = Source.objects.get(pk=source_id)
        action = source.get_action(name=action_name)
    except OperationalError as exception:
        raise self.retry(exc=exception)

    action.background_task(interface_load_kwargs=action_interface_kwargs)


@app.task(ignore_result=True)
def task_source_backend_action_execute(
    action_name, source_id, action_interface_kwargs=None, user_id=None
):
    # This task is not be retried because it runs on a schedule.
    # Retrying the task can cause the same source file to be uploaded twice.
    Source = apps.get_model(
        app_label='sources', model_name='Source'
    )

    User = get_user_model()

    action_interface_kwargs = action_interface_kwargs or {}

    lock_id = 'task_source_backend_action_default_execute-%d' % source_id

    try:
        logger.debug('trying to acquire lock: %s', lock_id)
        lock = LockingBackend.get_backend().acquire_lock(
            name=lock_id, timeout=DEFAULT_SOURCES_LOCK_EXPIRE
        )
    except LockError:
        logger.debug('unable to obtain lock: %s' % lock_id)
    else:
        logger.debug('acquired lock: %s', lock_id)

        try:
            source = Source.objects.get(pk=source_id)

            if user_id:
                user = User.objects.get(pk=user_id)
            else:
                user = None

            action_interface_kwargs['user'] = user

            action = source.get_action(name=action_name)

            action.execute(
                interface_name='Model',
                interface_load_kwargs=action_interface_kwargs
            )
        except Exception as exception:
            logger.error(
                'Error processing source id: %s; %s', source_id, exception,
                exc_info=True
            )
            source.error_log.create(
                text='{}; {}'.format(
                    exception.__class__.__name__, exception
                )
            )
            if settings.DEBUG:
                raise
        else:
            source.error_log.all().delete()
        finally:
            lock.release()



from django.apps import apps
from django.utils.translation import gettext_lazy as _

from mayan.apps.acls.classes import ModelPermission
from mayan.apps.common.apps import MayanAppConfig
from mayan.apps.common.menus import (
    menu_list_facet, menu_object, menu_secondary
)
from mayan.apps.events.classes import EventModelRegistry, ModelEventType
from mayan.apps.navigation.classes import SourceColumn

from .events import (
    event_document_comment_created, event_document_comment_deleted,
    event_document_comment_edited
)
from .links import (
    link_comment_add, link_comment_delete, link_comment_edit,
    link_comments_for_document
)
from .permissions import (
    permission_document_comment_create, permission_document_comment_delete,
    permission_document_comment_edit, permission_document_comment_view
)


class DocumentCommentsApp(MayanAppConfig):
    app_namespace = 'comments'
    app_url = 'comments'
    has_rest_api = True
    has_tests = True
    name = 'mayan.apps.document_comments'
    verbose_name = _(message='Document comments')

    def ready(self):
        super().ready()

        Document = apps.get_model(
            app_label='documents', model_name='Document'
        )

        Comment = self.get_model(model_name='Comment')

        EventModelRegistry.register(model=Comment)

        ModelEventType.register(
            model=Comment, event_types=(
                event_document_comment_edited,
            )
        )
        ModelEventType.register(
            model=Document, event_types=(
                event_document_comment_created,
                event_document_comment_deleted,
                event_document_comment_edited
            )
        )

        ModelPermission.register_inheritance(
            model=Comment, related='document',
        )
        ModelPermission.register(
            model=Document, permissions=(
                permission_document_comment_create,
                permission_document_comment_delete,
                permission_document_comment_edit,
                permission_document_comment_view
            )
        )

        SourceColumn(
            attribute='submit_date', is_identifier=True, is_sortable=True,
            source=Comment
        )
        SourceColumn(
            attribute='get_user_label', is_sortable=True,
            include_label=True, sort_field='user', source=Comment
        )
        SourceColumn(attribute='text', include_label=True, source=Comment)

        menu_list_facet.bind_links(
            links=(link_comments_for_document,), sources=(Document,)
        )

        menu_secondary.bind_links(
            links=(link_comment_add,),
            sources=(
                'comments:comments_for_document', 'comments:comment_add',
                'comments:comment_delete', 'comments:comment_details',
                'comments:comment_edit', 'comments:comment_multiple_delete'
            )
        )

        menu_object.bind_links(
            links=(
                link_comment_delete, link_comment_edit
            ), sources=(Comment,)
        )



import logging

from django.conf import settings
from django.contrib import messages
from django.core.files import File
from django.utils.translation import gettext_lazy as _

from mayan.apps.sources.literals import STORAGE_NAME_SOURCE_CACHE_FOLDER
from mayan.apps.storage.classes import DefinedStorage

from ..classes import SourceStoredFile
from ..forms import StoredFileUploadForm
from ..source_backend_actions.stored_file_actions import (
    SourceBackendActionFileStoredDeleteInteractive,
    SourceBackendActionFileStoredDeleteInteractiveNot,
    SourceBackendActionFileStoredDocumentFileUpload,
    SourceBackendActionFileStoredDocumentUpload,
    SourceBackendActionFileStoredImage, SourceBackendActionFileStoredList
)
from ..views import SourceBackendStoredFileSourceFileListView

logger = logging.getLogger(name=__name__)


class SourceBackendMixinStoredFileUploadBase:
    upload_form_class = StoredFileUploadForm

    @classmethod
    def get_form_fields(cls):
        fields = super().get_form_fields()

        fields.update(
            {
                'delete_after_upload': {
                    'class': 'django.forms.BooleanField',
                    'help_text': _(
                        'Delete the file after is has been successfully '
                        'uploaded.'
                    ),
                    'label': _(message='Delete after upload'),
                    'required': False
                }
            }
        )

        return fields

    @classmethod
    def get_form_fieldsets(cls):
        fieldsets = super().get_form_fieldsets()

        fieldsets += (
            (
                _(message='File content'), {
                    'fields': (
                        'delete_after_upload',
                    )
                },
            ),
        )

        return fieldsets

    def get_view_context(self, context, request):
        try:
            source_stored_file_list = list(
                self.get_stored_file_list()
            )
        except Exception as exception:
            messages.error(
                message=_(
                    'Unable get list of files; %s'
                ) % exception, request=request
            )
            source_stored_file_list = ()
            if settings.DEBUG or settings.TESTING:
                raise

        # Instantiate a fake list view to populate the pagination data for
        # the staging source file list.
        view = SourceBackendStoredFileSourceFileListView()
        view.kwargs = self.kwargs
        view.object_list = source_stored_file_list
        view.request = request

        template_source_stored_file_list_context = {
            'hide_link': True,
            'no_results_icon': self.icon,
            'no_results_text': _(
                'This could mean that the source file list is empty. It '
                'could also mean that the operating system user account '
                'being used for Mayan EDMS doesn\'t have the necessary '
                'file system permissions to access the source files.'
            ),
            'no_results_title': _(message='No source files available')
        }

        template_source_stored_file_list_context.update(
            view.get_context_data()
        )

        subtemplates_list = [
            {
                'context': {
                    'forms': context['forms']
                },
                'name': 'appearance/partials/form/multiple.html'
            },
            {
                'context': template_source_stored_file_list_context,
                'name': 'appearance/partials/list/table.html'
            }
        ]

        return {
            'subtemplates_list': subtemplates_list
        }


class SourceBackendMixinStoredFileDocumentFileUpload(
    SourceBackendMixinStoredFileUploadBase
):
    def get_action_class_list(self):
        action_class_list = super().get_action_class_list()

        action_class_list += (
            SourceBackendActionFileStoredDocumentFileUpload,
        )

        return action_class_list


class SourceBackendMixinStoredFileDocumentUpload(
    SourceBackendMixinStoredFileUploadBase
):
    def get_action_class_list(self):
        action_class_list = super().get_action_class_list()

        action_class_list += (SourceBackendActionFileStoredDocumentUpload,)

        return action_class_list


class SourceBackendMixinStoredFileDeleteInteractive:
    def action_file_delete(self, encoded_filename):
        source_stored_file = self.get_stored_file(
            encoded_filename=encoded_filename
        )

        source_stored_file.delete()

    def get_action_class_list(self):
        action_class_list = super().get_action_class_list()

        action_class_list += (
            SourceBackendActionFileStoredDeleteInteractive,
        )

        return action_class_list


class SourceBackendMixinStoredFileDeleteInteractiveNot:
    def action_file_delete(self, encoded_filename):
        source_stored_file = self.get_stored_file(
            encoded_filename=encoded_filename
        )

        source_stored_file.delete()

    def get_action_class_list(self):
        action_class_list = super().get_action_class_list()

        action_class_list += (
            SourceBackendActionFileStoredDeleteInteractiveNot,
        )

        return action_class_list


class SourceBackendMixinStoredFileImage:
    @classmethod
    def get_form_fields(cls):
        fields = super().get_form_fields()

        fields.update(
            {
                'preview_width': {
                    'class': 'django.forms.IntegerField',
                    'help_text': _(
                        'Width value to be passed to the converter backend.'
                    ),
                    'kwargs': {
                        'min_value': 0
                    },
                    'label': _(message='Preview width'),
                    'required': True
                },
                'preview_height': {
                    'class': 'django.forms.IntegerField',
                    'help_text': _(
                        'Height value to be passed to the converter backend.'
                    ),
                    'kwargs': {
                        'min_value': 0
                    },
                    'label': _(message='Preview height'),
                    'required': False
                },
                'preview_max_size': {
                    'class': 'django.forms.IntegerField',
                    'default': SourceStoredFile.DEFAULT_PREVIEW_MAX_SIZE,
                    'help_text': _(
                        'Maximum size in bytes for which previews will be '
                        'generated. Example: 20000 = 20KB.'
                    ),
                    'label': _(message='Max size preview'),
                    'required': True
                }
            }
        )

        return fields

    @classmethod
    def get_form_fieldsets(cls):
        fieldsets = super().get_form_fieldsets()

        fieldsets += (
            (
                _(message='File images'), {
                    'fields': (
                        'preview_width', 'preview_height',
                        'preview_max_size'
                    )
                },
            ),
        )

        return fieldsets

    def action_file_image(
        self, encoded_filename, maximum_layer_order=None,
        transformation_instance_list=None, user=None
    ):
        if maximum_layer_order:
            maximum_layer_order = int(maximum_layer_order)

        source_stored_file = self.get_stored_file(
            encoded_filename=encoded_filename
        )

        combined_transformation_list = source_stored_file.get_combined_transformation_list(
            maximum_layer_order=maximum_layer_order,
            transformation_instance_list=transformation_instance_list,
            user=user
        )

        cache_filename = source_stored_file.generate_image(
            transformation_instance_list=combined_transformation_list
        )

        storage_source_stored_file_image_cache = DefinedStorage.get(
            name=STORAGE_NAME_SOURCE_CACHE_FOLDER
        ).get_storage_instance()

        def file_generator():
            with storage_source_stored_file_image_cache.open(name=cache_filename) as file_object:
                while True:
                    chunk = file_object.read(File.DEFAULT_CHUNK_SIZE)
                    if not chunk:
                        break
                    else:
                        yield chunk

        return file_generator

    def get_action_class_list(self):
        action_class_list = super().get_action_class_list()

        action_class_list += (SourceBackendActionFileStoredImage,)

        return action_class_list


class SourceBackendMixinStoredFileList:
    def action_file_list(self):
        for source_stored_file in self.get_stored_file_list():
            yield {
                'encoded_filename': source_stored_file.encoded_filename,
                'filename': source_stored_file.filename,
                'size': source_stored_file.get_size()
            }

    def get_action_class_list(self):
        action_class_list = super().get_action_class_list()

        action_class_list += (SourceBackendActionFileStoredList,)

        return action_class_list



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('document_states', '0020_auto_20200311_1714')
    ]

    operations = [
        migrations.AlterField(
            model_name='workflowstateaction',
            name='condition',
            field=models.TextField(
                blank=True, help_text="The condition that will determine "
                "if this state action is executed or not. The condition "
                "is evaluated against the workflow instance. Conditions "
                "that do not return any value, that return the Python "
                "logical None, or an empty string ('') are considered "
                "to be logical false, any other value is considered to "
                "be the logical true.", verbose_name='Condition'
            )
        ),
        migrations.AlterField(
            model_name='workflowtransition',
            name='condition',
            field=models.TextField(
                blank=True, help_text="The condition that will determine "
                "if this transition is enabled or not. The condition is "
                "evaluated against the workflow instance. Conditions that "
                "do not return any value, that return the Python logical "
                "None, or an empty string ('') are considered to be "
                "logical false, any other value is considered to be the "
                "logical true.", verbose_name='Condition'
            )
        )
    ]



from django.db import migrations


def code_document_latest_file_populate(apps, schema_editor):
    Document = apps.get_model(
        app_label='documents', model_name='Document'
    )

    for document in Document.objects.all():
        document.file_latest = document.files.order_by('timestamp').only('id').last()
        document.save(
            update_fields=('file_latest',)
        )


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0085_document_file_latest')
    ]

    operations = [
        migrations.RunPython(
            code=code_document_latest_file_populate,
            reverse_code=migrations.RunPython.noop
        )
    ]



from .workflow_template_transition_mixins import WorkflowTemplateTransitionTestMixin


class WorkflowToolViewTestMixin(WorkflowTemplateTransitionTestMixin):
    def _request_workflow_launch_view(self):
        return self.post(
            viewname='document_states:tool_launch_workflows',
        )



from unittest import mock

from mayan.apps.cabinets.tests.mixins import CabinetTestMixin
from mayan.apps.common.tests.mixins import ManagementCommandTestMixin
from mayan.apps.document_indexing.tests.mixins import IndexTemplateTestMixin
from mayan.apps.storage.utils import NamedTemporaryFile
from mayan.apps.testing.tests.base import BaseTestCase

from ..literals import (
    COMMAND_NAME_MIRRORING_MOUNT_CABINET, COMMAND_NAME_MIRRORING_MOUNT_INDEX
)


class MountCabinetManagementCommandTestCase(
    ManagementCommandTestMixin, CabinetTestMixin, BaseTestCase
):
    _test_management_command_name = COMMAND_NAME_MIRRORING_MOUNT_CABINET
    auto_create_test_cabinet = True

    @mock.patch('fuse.FUSE', autospec=True)
    def test_cabinet_mount_basic(self, FUSE):
        with NamedTemporaryFile() as file_object:
            FUSE.return_value = None
            self._call_test_management_command(
                label=self._test_cabinet.label, mount_point=file_object.name
            )

    @mock.patch('fuse.FUSE', autospec=True)
    def test_cabinet_mount_with_invalid_slug(self, FUSE):
        with NamedTemporaryFile() as file_object:
            FUSE.return_value = None
            with self.assertRaises(expected_exception=SystemExit):
                self._call_test_management_command(
                    label='invalid', mount_point=file_object.name
                )


class MountIndexManagementCommandTestCase(
    IndexTemplateTestMixin, ManagementCommandTestMixin, BaseTestCase
):
    _test_management_command_name = COMMAND_NAME_MIRRORING_MOUNT_INDEX
    auto_create_test_index_template = False

    def setUp(self):
        super().setUp()
        self._create_test_index_template()

    @mock.patch('fuse.FUSE', autospec=True)
    def test_index_mount_basic(self, FUSE):
        with NamedTemporaryFile() as file_object:
            FUSE.return_value = None
            self._call_test_management_command(
                mount_point=file_object.name,
                slug=self._test_index_template.slug
            )

    @mock.patch('fuse.FUSE', autospec=True)
    def test_index_mount_with_invalid_slug(self, FUSE):
        with NamedTemporaryFile() as file_object:
            FUSE.return_value = None
            with self.assertRaises(expected_exception=SystemExit):
                self._call_test_management_command(
                    mount_point=file_object.name, slug='invalid'
                )



from django.utils.translation import gettext_lazy as _

from mayan.apps.permissions.classes import PermissionNamespace

namespace = PermissionNamespace(
    label=_(message='Document parsing'), name='document_parsing'
)

permission_document_file_content_view = namespace.add_permission(
    label=_(message='View the content of a document file'), name='content_view'
)
permission_document_file_parse = namespace.add_permission(
    label=_(message='Parse the content of a document file'), name='parse_document'
)
permission_document_type_parsing_setup = namespace.add_permission(
    label=_(message='Change document type parsing settings'),
    name='document_type_setup'
)



from django.urls import re_path

from .api_views import (
    APIACLDetailView, APIACLListView, APIACLPermissionAddView,
    APIACLPermissionListView, APIACLPermissionRemoveView,
    APIClassPermissionList
)
from .views import (
    ACLCreateView, ACLDeleteView, ACLListView, ACLPermissionAddRemoveView,
    GlobalACLListView
)

urlpatterns = [
    re_path(
        route=r'^acls/global/', name='global_acl_list',
        view=GlobalACLListView.as_view()
    ),
    re_path(
        route=r'^acls/(?P<acl_id>\d+)/delete/$', name='acl_delete',
        view=ACLDeleteView.as_view()
    ),
    re_path(
        route=r'^acls/(?P<acl_id>\d+)/permissions/$', name='acl_permissions',
        view=ACLPermissionAddRemoveView.as_view()
    ),
    re_path(
        route=r'^apps/(?P<app_label>[-\w]+)/models/(?P<model_name>[-\w]+)/objects/(?P<object_id>\d+)/acls/$',
        name='acl_list', view=ACLListView.as_view()
    ),
    re_path(
        route=r'^apps/(?P<app_label>[-\w]+)/models/(?P<model_name>[-\w]+)/objects/(?P<object_id>\d+)/acls/create/$',
        name='acl_create', view=ACLCreateView.as_view()
    )
]

api_urls = [
    re_path(
        route=r'^objects/(?P<app_label>[-\w]+)/(?P<model_name>[-\w]+)/permissions/$',
        name='class-permission-list', view=APIClassPermissionList.as_view()
    ),
    re_path(
        route=r'^objects/(?P<app_label>[-\w]+)/(?P<model_name>[-\w]+)/(?P<object_id>\d+)/acls/$',
        name='accesscontrollist-list', view=APIACLListView.as_view()
    ),
    re_path(
        route=r'^objects/(?P<app_label>[-\w]+)/(?P<model_name>[-\w]+)/(?P<object_id>\d+)/acls/(?P<acl_id>\d+)/$',
        name='accesscontrollist-detail', view=APIACLDetailView.as_view()
    ),
    re_path(
        route=r'^objects/(?P<app_label>[-\w]+)/(?P<model_name>[-\w]+)/(?P<object_id>\d+)/acls/(?P<acl_id>\d+)/permissions/add/$',
        name='accesscontrollist-permission-add',
        view=APIACLPermissionAddView.as_view()
    ),
    re_path(
        route=r'^objects/(?P<app_label>[-\w]+)/(?P<model_name>[-\w]+)/(?P<object_id>\d+)/acls/(?P<acl_id>\d+)/permissions/$',
        name='accesscontrollist-permission-list',
        view=APIACLPermissionListView.as_view()
    ),
    re_path(
        route=r'^objects/(?P<app_label>[-\w]+)/(?P<model_name>[-\w]+)/(?P<object_id>\d+)/acls/(?P<acl_id>\d+)/permissions/remove/$',
        name='accesscontrollist-permission-remove',
        view=APIACLPermissionRemoveView.as_view()
    )
]



from django.utils.translation import gettext_lazy as _

from .classes import ErrorLog as ErrorLogProxy


class ErrorLogPartitionEntryBusinessLogicMixin:
    def get_object(self):
        return self.error_log_partition.content_object

    get_object.short_description = _(message='Object')


class StoredErrorLogBusinessLogicMixin:
    @property
    def app_label(self):
        return self.proxy.app_config

    app_label.fget.short_description = _(message='App label')

    @property
    def proxy(self):
        return ErrorLogProxy.get(name=self.name)



from mayan.apps.testing.tests.base import BaseTestCase

from .literals import (
    TEST_BOOTSTAP_SETTING_NAME, TEST_CONFIG_FILE_NAME,
    TEST_CONFIG_FILE_VALUE, TEST_ENVIRONMENT_VALUE,
    TEST_ENVIRONMENT_VARIABLE_NAME, TEST_GLOBAL_NAME, TEST_GLOBAL_VALUE,
    TEST_SETTING_VALUE, TEST_SETTING_VALUE_OVERRIDE
)
from .mixins import BoostrapSettingTestMixin


class BoostrapSettingTestCase(BoostrapSettingTestMixin, BaseTestCase):
    def setUp(self):
        super().setUp()
        self._register_test_boostrap_setting()
        self._create_test_bootstrap_singleton()

    def test_bootstrap_environment_overrides_config(self):
        self._set_environment_variable(
            name='MAYAN_{}'.format(TEST_BOOTSTAP_SETTING_NAME),
            value=TEST_SETTING_VALUE_OVERRIDE
        )

        self._test_setting_global_name = TEST_BOOTSTAP_SETTING_NAME
        self._test_configuration_value = TEST_SETTING_VALUE
        self._create_test_configuration_file(
            callback=self._test_setting_namespace_singleton.update_globals
        )

        self.assertEqual(
            self.test_globals[TEST_BOOTSTAP_SETTING_NAME],
            TEST_SETTING_VALUE_OVERRIDE
        )

    def test_bootstrap_config_overrides_settings(self):
        self.test_globals[TEST_BOOTSTAP_SETTING_NAME] = TEST_SETTING_VALUE

        self._test_setting_global_name = TEST_BOOTSTAP_SETTING_NAME
        self._test_configuration_value = TEST_SETTING_VALUE_OVERRIDE
        self._create_test_configuration_file(
            callback=self._test_setting_namespace_singleton.update_globals
        )

        self.assertEqual(
            self.test_globals[TEST_BOOTSTAP_SETTING_NAME],
            TEST_SETTING_VALUE_OVERRIDE
        )

    def test_bootstrap_settings_overrides_default(self):
        self.test_globals[
            TEST_BOOTSTAP_SETTING_NAME
        ] = TEST_SETTING_VALUE_OVERRIDE

        self._test_setting_namespace_singleton.update_globals()

        self.assertEqual(
            self.test_globals[TEST_BOOTSTAP_SETTING_NAME],
            TEST_SETTING_VALUE_OVERRIDE
        )

    def test_bootstrap_default(self):
        self._test_setting_namespace_singleton.update_globals()

        self.assertEqual(
            self.test_globals[TEST_BOOTSTAP_SETTING_NAME], 'value default'
        )


class BoostrapSettingTemplateTestCase(
    BoostrapSettingTestMixin, BaseTestCase
):
    def setUp(self):
        super().setUp()
        self._register_test_boostrap_setting()
        self._create_test_bootstrap_singleton()

    def test_bootstrap_environment_template(self):
        self.test_globals[TEST_GLOBAL_NAME] = TEST_GLOBAL_VALUE

        self._set_environment_variable(
            name=TEST_ENVIRONMENT_VARIABLE_NAME,
            value=TEST_ENVIRONMENT_VALUE
        )
        self._set_environment_variable(
            name='MAYAN_SETTING_TEMPLATE_{}'.format(TEST_BOOTSTAP_SETTING_NAME),
            value='{{{{ {} }}}}-{{{{ {} }}}}-{{{{ {} }}}}'.format(
                TEST_GLOBAL_NAME, TEST_CONFIG_FILE_NAME,
                TEST_ENVIRONMENT_VARIABLE_NAME
            )
        )

        self._test_setting_global_name = TEST_CONFIG_FILE_NAME
        self._test_configuration_value = TEST_CONFIG_FILE_VALUE
        self._create_test_configuration_file(
            callback=self._test_setting_namespace_singleton.update_globals
        )

        self.assertEqual(
            self.test_globals[TEST_BOOTSTAP_SETTING_NAME],
            '{}-{}-{}'.format(
                TEST_GLOBAL_VALUE, TEST_CONFIG_FILE_VALUE,
                TEST_ENVIRONMENT_VALUE
            )
        )

    def test_bootstrap_global_template(self):
        self.test_globals[TEST_GLOBAL_NAME] = TEST_GLOBAL_VALUE

        self.test_globals[
            'SETTING_TEMPLATE_{}'.format(TEST_BOOTSTAP_SETTING_NAME)
        ] = '{{{{ {} }}}}-{{{{ {} }}}}-{{{{ {} }}}}'.format(
            TEST_GLOBAL_NAME, TEST_CONFIG_FILE_NAME,
            TEST_ENVIRONMENT_VARIABLE_NAME
        )

        self._set_environment_variable(
            name=TEST_ENVIRONMENT_VARIABLE_NAME,
            value=TEST_ENVIRONMENT_VALUE
        )

        self._test_setting_global_name = TEST_CONFIG_FILE_NAME
        self._test_configuration_value = TEST_CONFIG_FILE_VALUE
        self._create_test_configuration_file(
            callback=self._test_setting_namespace_singleton.update_globals
        )

        self.assertEqual(
            self.test_globals[TEST_BOOTSTAP_SETTING_NAME],
            '{}-{}-{}'.format(
                TEST_GLOBAL_VALUE, TEST_CONFIG_FILE_VALUE,
                TEST_ENVIRONMENT_VALUE
            )
        )

    def test_bootstrap_environment_template_overrides(self):
        self.test_globals[TEST_BOOTSTAP_SETTING_NAME] = TEST_SETTING_VALUE

        self._set_environment_variable(
            name='MAYAN_{}'.format(TEST_BOOTSTAP_SETTING_NAME),
            value=TEST_SETTING_VALUE
        )

        self._set_environment_variable(
            name='MAYAN_SETTING_TEMPLATE_{}'.format(
                TEST_BOOTSTAP_SETTING_NAME
            ), value=TEST_SETTING_VALUE_OVERRIDE
        )

        self._test_setting_global_name = TEST_BOOTSTAP_SETTING_NAME
        self._test_configuration_value = TEST_SETTING_VALUE
        self._create_test_configuration_file(
            callback=self._test_setting_namespace_singleton.update_globals
        )

        self.assertEqual(
            self.test_globals[TEST_BOOTSTAP_SETTING_NAME],
            TEST_SETTING_VALUE_OVERRIDE
        )



from mayan.apps.navigation.column_widgets import SourceColumnWidget

from .icons import icon_fail as default_icon_fail, icon_ok as default_icon_ok


class TwoStateWidget(SourceColumnWidget):
    template_name = 'views/two_state_widget.html'

    def __init__(self, icon_ok=None, icon_fail=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.icon_ok = icon_ok or default_icon_ok
        self.icon_fail = icon_fail or default_icon_fail

    def get_extra_context(self):
        return {
            'icon_fail': self.icon_fail, 'icon_ok': self.icon_ok
        }


class ObjectLinkWidget(SourceColumnWidget):
    template_string = '<a href="{{ url }}">{{ object_type }}{{ label }}</a>'

    def get_extra_context(self):
        label = ''
        object_type = ''
        url = None

        if self.value:
            label = str(self.value)
            object_type = '{}: '.format(self.value._meta.verbose_name)
            try:
                url = self.value.get_absolute_url()
            except AttributeError:
                url = None

            if getattr(self.value, 'is_staff', None) or getattr(self.value, 'is_superuser', None):
                # Don't display a anchor to for the user details view for
                # super users and staff, the details view filters them.
                # Staff and admin users are not manageable by the normal
                # user views.
                url = '#'
                label = '{}{}'.format(object_type, label)

        return {
            'label': label, 'object_type': object_type,
            'url': url or '#'
        }



from mayan.apps.appearance.classes import Icon

icon_check_in_document = Icon(
    driver_name='fontawesome-dual', primary_symbol='shopping-cart',
    secondary_symbol='minus'
)
icon_check_out_document = Icon(
    driver_name='fontawesome-dual', primary_symbol='shopping-cart',
    secondary_symbol='plus'
)
icon_check_out_info = Icon(driver_name='fontawesome', symbol='shopping-cart')
icon_dashboard_check_outs = Icon(
    driver_name='fontawesome', symbol='shopping-cart'
)
icon_check_out_list = Icon(
    driver_name='fontawesome', symbol='shopping-cart'
)



from django.db.models import Q

from ..classes import Permission, PermissionNamespace
from ..models import Role

from .literals import (
    TEST_CASE_ROLE_LABEL, TEST_PERMISSION_LABEL, TEST_PERMISSION_NAME,
    TEST_PERMISSION_NAMESPACE_LABEL, TEST_PERMISSION_NAMESPACE_NAME,
    TEST_ROLE_LABEL, TEST_ROLE_LABEL_EDITED
)


class PermissionAPIViewTestMixin:
    def _request_permissions_list_api_view(self):
        return self.get(viewname='rest_api:permission-list')


class PermissionTestMixin:
    def setUp(self):
        super().setUp()
        self._test_permission_list = []

    def _create_test_permission(self):
        self._test_permission_namespace = PermissionNamespace(
            label=TEST_PERMISSION_NAMESPACE_LABEL,
            name=TEST_PERMISSION_NAMESPACE_NAME
        )

        test_permission_count = len(self._test_permission_list)
        self._test_permission = self._test_permission_namespace.add_permission(
            label='{}_{}'.format(
                TEST_PERMISSION_LABEL, test_permission_count
            ), name='{}_{}'.format(
                TEST_PERMISSION_NAME, test_permission_count
            )
        )
        self._test_permission_list.append(self._test_permission)


class PermissionTestCaseMixin:
    def setUp(self):
        super().setUp()
        Permission.invalidate_cache()


class RoleTestMixin:
    def setUp(self):
        super().setUp()
        self._test_roles = []

    def _create_test_role(self, add_groups=None):
        total_test_roles = len(self._test_roles)
        label = '{}_{}'.format(TEST_ROLE_LABEL, total_test_roles)

        self._test_role = Role.objects.create(label=label)

        self._test_roles.append(self._test_role)

        for group in add_groups or []:
            self._test_role.groups.add(group)


class GroupRoleAddRemoveViewTestMixin(RoleTestMixin):
    def _request_test_group_role_add_remove_get_view(self):
        return self.get(
            viewname='permissions:group_role_list', kwargs={
                'group_id': self._test_group.pk
            }
        )

    def _request_test_group_role_add_view(self):
        return self.post(
            viewname='permissions:group_role_list', kwargs={
                'group_id': self._test_group.pk,
            }, data={
                'available-selection': self._test_role.pk,
                'available-submit': 'true'
            }
        )

    def _request_test_group_role_remove_view(self):
        return self.post(
            viewname='permissions:group_role_list', kwargs={
                'group_id': self._test_group.pk,
            }, data={
                'added-selection': self._test_role.pk,
                'added-submit': 'true'
            }
        )


class RoleAPIViewTestMixin(RoleTestMixin):
    def _request_test_role_create_api_view(self, extra_data=None):
        pk_list = list(
            Role.objects.values_list('pk', flat=True)
        )

        data = {'label': TEST_ROLE_LABEL}

        if extra_data:
            data.update(extra_data)

        response = self.post(
            viewname='rest_api:role-list', data=data
        )

        try:
            self._test_role = Role.objects.get(
                ~Q(pk__in=pk_list)
            )
        except Role.DoesNotExist:
            self._test_role = None

        return response

    def _request_test_role_delete_api_view(self):
        return self.delete(
            viewname='rest_api:role-detail', kwargs={
                'role_id': self._test_role.pk
            }
        )

    def _request_test_role_edit_api_view(
        self, extra_data=None, request_type='patch'
    ):
        data = {'label': TEST_ROLE_LABEL_EDITED}

        if extra_data:
            data.update(extra_data)

        return getattr(self, request_type)(
            viewname='rest_api:role-detail', kwargs={
                'role_id': self._test_role.pk
            }, data=data
        )

    def _request_test_role_edit_api_patch_view_extra_data(self):
        extra_data = {
            'groups_pk_list': '{}'.format(self._test_group.pk),
            'permissions_pk_list': '{}'.format(self._test_permission.pk)
        }
        return self._request_test_role_edit_api_view(
            extra_data=extra_data, request_type='patch'
        )

    def _request_test_role_edit_api_put_view_extra_data(self):
        extra_data = {
            'groups_pk_list': '{}'.format(self._test_group.pk),
            'permissions_pk_list': '{}'.format(self._test_permission.pk)
        }
        return self._request_test_role_edit_api_view(
            extra_data=extra_data, request_type='put'
        )

    def _request_test_role_list_api_view(self):
        return self.get(viewname='rest_api:role-list')


class RoleGroupAPIViewTestMixin(RoleTestMixin):
    def _request_test_role_group_add_api_view(self):
        return self.post(
            viewname='rest_api:role-group-add', kwargs={
                'role_id': self._test_role.pk
            }, data={'group': self._test_group.pk}
        )

    def _request_test_role_group_list_api_view(self):
        return self.get(
            viewname='rest_api:role-group-list', kwargs={
                'role_id': self._test_role.pk
            }
        )

    def _request_test_role_group_remove_api_view(self):
        return self.post(
            viewname='rest_api:role-group-remove', kwargs={
                'role_id': self._test_role.pk
            }, data={'group': self._test_group.pk}
        )


class RolePermissionAPIViewTestMixin(RoleTestMixin):
    def _request_test_role_permission_add_api_view(self):
        return self.post(
            viewname='rest_api:role-permission-add', kwargs={
                'role_id': self._test_role.pk
            }, data={'permission': self._test_permission.pk}
        )

    def _request_test_role_permission_list_api_view(self):
        return self.get(
            viewname='rest_api:role-permission-list', kwargs={
                'role_id': self._test_role.pk
            }
        )

    def _request_test_role_permission_remove_api_view(self):
        return self.post(
            viewname='rest_api:role-permission-remove', kwargs={
                'role_id': self._test_role.pk
            }, data={'permission': self._test_permission.pk}
        )


class RoleTestCaseMixin:
    def setUp(self):
        super().setUp()
        if hasattr(self, '_test_case_group'):
            self.create_role()

    def create_role(self):
        self._test_case_role = Role.objects.create(
            label=TEST_CASE_ROLE_LABEL
        )

    def grant_permission(self, permission):
        self._test_case_role.grant(permission=permission)

    def revoke_permission(self, permission):
        self._test_case_role.revoke(permission=permission)


class RoleGroupAddRemoveViewTestMixin(RoleTestMixin):
    def _request_test_role_group_add_remove_get_view(self):
        return self.get(
            viewname='permissions:role_group_list', kwargs={
                'role_id': self._test_role.pk
            }
        )

    def _request_test_role_group_add_view(self):
        return self.post(
            viewname='permissions:role_group_list', kwargs={
                'role_id': self._test_role.pk
            }, data={
                'available-submit': 'true',
                'available-selection': self._test_group.pk
            }
        )

    def _request_test_role_group_remove_view(self):
        return self.post(
            viewname='permissions:role_group_list', kwargs={
                'role_id': self._test_role.pk
            }, data={
                'added-submit': 'true',
                'added-selection': self._test_group.pk
            }
        )


class RolePermissionAddRemoveViewTestMixin(RoleTestMixin):
    def _request_test_role_permission_add_remove_get_view(self):
        return self.get(
            viewname='permissions:role_permission_list', kwargs={
                'role_id': self._test_role.pk
            }
        )

    def _request_test_role_permission_add_view(self):
        return self.post(
            viewname='permissions:role_permission_list', kwargs={
                'role_id': self._test_role.pk
            }, data={
                'available-selection': self._test_permission.stored_permission.pk,
                'available-submit': 'true'
            }
        )

    def _request_test_role_permission_remove_view(self):
        return self.post(
            viewname='permissions:role_permission_list', kwargs={
                'role_id': self._test_role.pk
            }, data={
                'added-selection': self._test_permission.stored_permission.pk,
                'added-submit': 'true'
            }
        )


class RoleViewTestMixin(RoleTestMixin):
    def _request_test_role_create_view(self):
        pk_list = list(
            Role.objects.values_list('pk', flat=True)
        )

        response = self.post(
            viewname='permissions:role_create', data={
                'label': TEST_ROLE_LABEL
            }
        )

        try:
            self._test_role = Role.objects.get(
                ~Q(pk__in=pk_list)
            )
        except Role.DoesNotExist:
            self._test_role = None

        return response

    def _request_test_role_single_delete_view(self):
        return self.post(
            viewname='permissions:role_single_delete', kwargs={
                'role_id': self._test_role.pk
            }
        )

    def _request_test_role_multiple_delete_view(self):
        return self.post(
            viewname='permissions:role_multiple_delete', data={
                'id_list': self._test_role.pk
            }
        )

    def _request_test_role_edit_view(self):
        return self.post(
            viewname='permissions:role_edit', kwargs={
                'role_id': self._test_role.pk
            }, data={
                'label': TEST_ROLE_LABEL_EDITED
            }
        )

    def _request_test_role_list_view(self):
        return self.get(viewname='permissions:role_list')



from django.core import management

from .literals import COMMAND_NAME_MIGRATE


def handler_pre_initial_setup(sender, **kwargs):
    management.call_command(
        command_name=COMMAND_NAME_MIGRATE, interactive=False
    )


def handler_pre_upgrade(sender, **kwargs):
    management.call_command(
        command_name=COMMAND_NAME_MIGRATE, fake_initial=True,
        interactive=False
    )



from django.conf import settings
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('locales', '0002_auto_20210130_0324'),
        ('common', '0018_delete_userlocaleprofile'),
    ]

    operations = [
        migrations.RenameModel(
            old_name='UserLocaleProfileNew',
            new_name='UserLocaleProfile',
        ),
    ]



import logging

from mayan.apps.documents.events import (
    event_document_created, event_document_file_created,
    event_document_file_edited, event_document_version_created,
    event_document_version_edited, event_document_version_page_created
)
from mayan.apps.documents.models.document_file_models import DocumentFile
from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.tests.base import GenericDocumentTestCase
from mayan.apps.user_management.events import (
    event_group_created, event_user_created
)
from mayan.apps.user_management.tests.mixins import GroupTestMixin

from ..events import event_quota_created
from ..exceptions import QuotaExceeded
from ..quota_backends import DocumentCountQuota, DocumentSizeQuota


class DocumentCountQuotaTestCase(GroupTestMixin, GenericDocumentTestCase):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        # Increase the initial usage count to 1 by uploading a document
        # as the test case user.
        self._clear_events()
        self._upload_test_document(user=self._test_case_user)
        self.test_case_silenced_logger_new_level = logging.FATAL + 10
        self._silence_logger(name='mayan.apps.documents.model_mixins')

    def test_user_all_document_type_all(self):
        self._test_quota = DocumentCountQuota.create(
            documents_limit=1,
            document_type_all=True,
            document_type_ids=(),
            group_ids=(),
            user_all=True,
            user_ids=()
        )

        with self.assertRaises(expected_exception=QuotaExceeded):
            self._upload_test_document()

        events = self._get_test_events()
        self.assertEqual(events.count(), 7)

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, self._test_document)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, self._test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, self._test_document)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, self._test_document_file)
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        self.assertEqual(events[3].action_object, self._test_document)
        self.assertEqual(events[3].actor, self._test_case_user)
        self.assertEqual(events[3].target, self._test_document_version)
        self.assertEqual(events[3].verb, event_document_version_created.id)

        self.assertEqual(
            events[4].action_object, self._test_document_version
        )
        self.assertEqual(events[4].actor, self._test_case_user)
        self.assertEqual(events[4].target, self._test_document_version_page)
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[5].action_object, self._test_document)
        self.assertEqual(events[5].actor, self._test_case_user)
        self.assertEqual(events[5].target, self._test_document_version)
        self.assertEqual(events[5].verb, event_document_version_edited.id)

        self.assertEqual(events[6].action_object, None)
        self.assertEqual(events[6].actor, self._test_quota)
        self.assertEqual(events[6].target, self._test_quota)
        self.assertEqual(events[6].verb, event_quota_created.id)

    def test_user_all_document_type_all_two_users(self):
        self._test_quota = DocumentCountQuota.create(
            documents_limit=1,
            document_type_all=True,
            document_type_ids=(),
            group_ids=(),
            user_all=True,
            user_ids=()
        )
        self._create_test_user()

        with self.assertRaises(expected_exception=QuotaExceeded):
            self._upload_test_document(user=self._test_user)

        events = self._get_test_events()
        self.assertEqual(events.count(), 8)

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, self._test_document)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, self._test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, self._test_document)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, self._test_document_file)
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        self.assertEqual(events[3].action_object, self._test_document)
        self.assertEqual(events[3].actor, self._test_case_user)
        self.assertEqual(events[3].target, self._test_document_version)
        self.assertEqual(events[3].verb, event_document_version_created.id)

        self.assertEqual(
            events[4].action_object, self._test_document_version
        )
        self.assertEqual(events[4].actor, self._test_case_user)
        self.assertEqual(events[4].target, self._test_document_version_page)
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[5].action_object, self._test_document)
        self.assertEqual(events[5].actor, self._test_case_user)
        self.assertEqual(events[5].target, self._test_document_version)
        self.assertEqual(events[5].verb, event_document_version_edited.id)

        self.assertEqual(events[6].action_object, None)
        self.assertEqual(events[6].actor, self._test_quota)
        self.assertEqual(events[6].target, self._test_quota)
        self.assertEqual(events[6].verb, event_quota_created.id)

        self.assertEqual(events[7].action_object, None)
        self.assertEqual(events[7].actor, self._test_user)
        self.assertEqual(events[7].target, self._test_user)
        self.assertEqual(events[7].verb, event_user_created.id)

    def test_user_all_document_type_test(self):
        self._test_quota = DocumentCountQuota.create(
            documents_limit=1,
            document_type_all=False,
            document_type_ids=(self._test_document_type.pk,),
            group_ids=(),
            user_all=True,
            user_ids=()
        )

        with self.assertRaises(expected_exception=QuotaExceeded):
            self._upload_test_document()

        events = self._get_test_events()
        self.assertEqual(events.count(), 7)

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, self._test_document)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, self._test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, self._test_document)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, self._test_document_file)
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        self.assertEqual(events[3].action_object, self._test_document)
        self.assertEqual(events[3].actor, self._test_case_user)
        self.assertEqual(events[3].target, self._test_document_version)
        self.assertEqual(events[3].verb, event_document_version_created.id)

        self.assertEqual(
            events[4].action_object, self._test_document_version
        )
        self.assertEqual(events[4].actor, self._test_case_user)
        self.assertEqual(events[4].target, self._test_document_version_page)
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[5].action_object, self._test_document)
        self.assertEqual(events[5].actor, self._test_case_user)
        self.assertEqual(events[5].target, self._test_document_version)
        self.assertEqual(events[5].verb, event_document_version_edited.id)

        self.assertEqual(events[6].action_object, None)
        self.assertEqual(events[6].actor, self._test_quota)
        self.assertEqual(events[6].target, self._test_quota)
        self.assertEqual(events[6].verb, event_quota_created.id)

    def test_user_test_document_type_all(self):
        self._test_quota = DocumentCountQuota.create(
            documents_limit=1,
            document_type_all=True,
            document_type_ids=(),
            group_ids=(),
            user_all=False,
            user_ids=(self._test_case_user.pk,)
        )

        with self.assertRaises(expected_exception=QuotaExceeded):
            self._upload_test_document(user=self._test_case_user)

        events = self._get_test_events()
        self.assertEqual(events.count(), 7)

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, self._test_document)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, self._test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, self._test_document)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, self._test_document_file)
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        self.assertEqual(events[3].action_object, self._test_document)
        self.assertEqual(events[3].actor, self._test_case_user)
        self.assertEqual(events[3].target, self._test_document_version)
        self.assertEqual(events[3].verb, event_document_version_created.id)

        self.assertEqual(
            events[4].action_object, self._test_document_version
        )
        self.assertEqual(events[4].actor, self._test_case_user)
        self.assertEqual(events[4].target, self._test_document_version_page)
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[5].action_object, self._test_document)
        self.assertEqual(events[5].actor, self._test_case_user)
        self.assertEqual(events[5].target, self._test_document_version)
        self.assertEqual(events[5].verb, event_document_version_edited.id)

        self.assertEqual(events[6].action_object, None)
        self.assertEqual(events[6].actor, self._test_quota)
        self.assertEqual(events[6].target, self._test_quota)
        self.assertEqual(events[6].verb, event_quota_created.id)

    def test_group_test_document_type_all(self):
        self._create_test_group()
        self._test_case_user.groups.add(self._test_group)

        self._test_quota = DocumentCountQuota.create(
            documents_limit=1,
            document_type_all=True,
            document_type_ids=(),
            group_ids=(self._test_group.pk,),
            user_all=False,
            user_ids=()
        )

        with self.assertRaises(expected_exception=QuotaExceeded):
            self._upload_test_document(user=self._test_case_user)

        events = self._get_test_events()
        self.assertEqual(events.count(), 8)

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, self._test_document)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, self._test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, self._test_document)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, self._test_document_file)
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        self.assertEqual(events[3].action_object, self._test_document)
        self.assertEqual(events[3].actor, self._test_case_user)
        self.assertEqual(events[3].target, self._test_document_version)
        self.assertEqual(events[3].verb, event_document_version_created.id)

        self.assertEqual(
            events[4].action_object, self._test_document_version
        )
        self.assertEqual(events[4].actor, self._test_case_user)
        self.assertEqual(events[4].target, self._test_document_version_page)
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[5].action_object, self._test_document)
        self.assertEqual(events[5].actor, self._test_case_user)
        self.assertEqual(events[5].target, self._test_document_version)
        self.assertEqual(events[5].verb, event_document_version_edited.id)

        self.assertEqual(events[6].action_object, None)
        self.assertEqual(events[6].actor, self._test_group)
        self.assertEqual(events[6].target, self._test_group)
        self.assertEqual(events[6].verb, event_group_created.id)

        self.assertEqual(events[7].action_object, None)
        self.assertEqual(events[7].actor, self._test_quota)
        self.assertEqual(events[7].target, self._test_quota)
        self.assertEqual(events[7].verb, event_quota_created.id)

    def test_allow(self):
        self._test_quota = DocumentCountQuota.create(
            documents_limit=1,
            document_type_all=True,
            document_type_ids=(),
            group_ids=(),
            user_all=False,
            user_ids=()
        )

        self._upload_test_document(user=self._test_case_user)

        events = self._get_test_events()
        self.assertEqual(events.count(), 13)

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_documents[0])
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, self._test_documents[0])
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, self._test_document_file_list[0])
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, self._test_documents[0])
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, self._test_document_file_list[0])
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        self.assertEqual(events[3].action_object, self._test_documents[0])
        self.assertEqual(events[3].actor, self._test_case_user)
        self.assertEqual(
            events[3].target, self._test_document_version_list[0]
        )
        self.assertEqual(events[3].verb, event_document_version_created.id)

        self.assertEqual(
            events[4].action_object, self._test_document_version_list[0]
        )
        self.assertEqual(events[4].actor, self._test_case_user)
        self.assertEqual(
            events[4].target, self._test_document_version_list[0].pages.first()
        )
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[5].action_object, self._test_documents[0])
        self.assertEqual(events[5].actor, self._test_case_user)
        self.assertEqual(
            events[5].target, self._test_document_version_list[0]
        )
        self.assertEqual(events[5].verb, event_document_version_edited.id)

        self.assertEqual(events[6].action_object, None)
        self.assertEqual(events[6].actor, self._test_quota)
        self.assertEqual(events[6].target, self._test_quota)
        self.assertEqual(events[6].verb, event_quota_created.id)

        self.assertEqual(events[7].action_object, self._test_document_type)
        self.assertEqual(events[7].actor, self._test_case_user)
        self.assertEqual(events[7].target, self._test_document)
        self.assertEqual(events[7].verb, event_document_created.id)

        self.assertEqual(events[8].action_object, self._test_document)
        self.assertEqual(events[8].actor, self._test_case_user)
        self.assertEqual(events[8].target, self._test_document_file)
        self.assertEqual(events[8].verb, event_document_file_created.id)

        self.assertEqual(events[9].action_object, self._test_document)
        self.assertEqual(events[9].actor, self._test_case_user)
        self.assertEqual(events[9].target, self._test_document_file)
        self.assertEqual(events[9].verb, event_document_file_edited.id)

        self.assertEqual(events[10].action_object, self._test_document)
        self.assertEqual(events[10].actor, self._test_case_user)
        self.assertEqual(events[10].target, self._test_document_version)
        self.assertEqual(events[10].verb, event_document_version_created.id)

        self.assertEqual(
            events[11].action_object, self._test_document_version
        )
        self.assertEqual(events[11].actor, self._test_case_user)
        self.assertEqual(events[11].target, self._test_document_version_page)
        self.assertEqual(
            events[11].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[12].action_object, self._test_document)
        self.assertEqual(events[12].actor, self._test_case_user)
        self.assertEqual(events[12].target, self._test_document_version)
        self.assertEqual(events[12].verb, event_document_version_edited.id)

    def test_super_user_restriction(self):
        self._create_test_super_user()

        self._test_quota = DocumentCountQuota.create(
            documents_limit=1,
            document_type_all=True,
            document_type_ids=(),
            group_ids=(),
            user_all=True,
            user_ids=()
        )

        self._upload_test_document(user=self._test_super_user)

        events = self._get_test_events()
        self.assertEqual(events.count(), 14)

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_documents[0])
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, self._test_documents[0])
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, self._test_document_file_list[0])
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, self._test_documents[0])
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, self._test_document_file_list[0])
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        self.assertEqual(events[3].action_object, self._test_documents[0])
        self.assertEqual(events[3].actor, self._test_case_user)
        self.assertEqual(
            events[3].target, self._test_document_version_list[0]
        )
        self.assertEqual(events[3].verb, event_document_version_created.id)

        self.assertEqual(
            events[4].action_object, self._test_document_version_list[0]
        )
        self.assertEqual(events[4].actor, self._test_case_user)
        self.assertEqual(
            events[4].target, self._test_document_version_list[0].pages.first()
        )
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[5].action_object, self._test_documents[0])
        self.assertEqual(events[5].actor, self._test_case_user)
        self.assertEqual(
            events[5].target, self._test_document_version_list[0]
        )
        self.assertEqual(events[5].verb, event_document_version_edited.id)

        self.assertEqual(events[6].action_object, None)
        self.assertEqual(events[6].actor, self._test_super_user)
        self.assertEqual(events[6].target, self._test_super_user)
        self.assertEqual(events[6].verb, event_user_created.id)

        self.assertEqual(events[7].action_object, None)
        self.assertEqual(events[7].actor, self._test_quota)
        self.assertEqual(events[7].target, self._test_quota)
        self.assertEqual(events[7].verb, event_quota_created.id)

        self.assertEqual(events[8].action_object, self._test_document_type)
        self.assertEqual(events[8].actor, self._test_super_user)
        self.assertEqual(events[8].target, self._test_document)
        self.assertEqual(events[8].verb, event_document_created.id)

        self.assertEqual(events[9].action_object, self._test_document)
        self.assertEqual(events[9].actor, self._test_super_user)
        self.assertEqual(events[9].target, self._test_document_file)
        self.assertEqual(events[9].verb, event_document_file_created.id)

        self.assertEqual(events[10].action_object, self._test_document)
        self.assertEqual(events[10].actor, self._test_super_user)
        self.assertEqual(events[10].target, self._test_document_file)
        self.assertEqual(events[10].verb, event_document_file_edited.id)

        self.assertEqual(events[11].action_object, self._test_document)
        self.assertEqual(events[11].actor, self._test_super_user)
        self.assertEqual(events[11].target, self._test_document_version)
        self.assertEqual(events[11].verb, event_document_version_created.id)

        self.assertEqual(
            events[12].action_object, self._test_document_version
        )
        self.assertEqual(events[12].actor, self._test_super_user)
        self.assertEqual(events[12].target, self._test_document_version_page)
        self.assertEqual(
            events[12].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[13].action_object, self._test_document)
        self.assertEqual(events[13].actor, self._test_super_user)
        self.assertEqual(events[13].target, self._test_document_version)
        self.assertEqual(events[13].verb, event_document_version_edited.id)


class DocumentSizeQuotaTestCase(GroupTestMixin, GenericDocumentTestCase):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self.test_case_silenced_logger_new_level = logging.FATAL + 10
        self._silence_logger(name='mayan.apps.documents.models')

    def test_user_all_document_type_all(self):
        self._test_quota = DocumentSizeQuota.create(
            document_size_limit=0.01,
            document_type_all=True,
            document_type_ids=(),
            group_ids=(),
            user_all=True,
            user_ids=()
        )

        test_document_count = Document.objects.count()
        test_document_file_count = DocumentFile.objects.count()

        self._clear_events()

        with self.assertRaises(expected_exception=QuotaExceeded):
            self._upload_test_document()

        self.assertEqual(
            Document.objects.count(), test_document_count + 1
        )
        self.assertEqual(
            DocumentFile.objects.count(), test_document_file_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        _test_document = Document.objects.first()
        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, _test_document)
        self.assertEqual(events[0].target, _test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

    def test_user_all_document_type_test(self):
        self._test_quota = DocumentSizeQuota.create(
            document_size_limit=0.01,
            document_type_all=False,
            document_type_ids=(self._test_document_type.pk,),
            group_ids=(),
            user_all=True,
            user_ids=()
        )

        test_document_count = Document.objects.count()
        test_document_file_count = DocumentFile.objects.count()

        self._clear_events()

        with self.assertRaises(expected_exception=QuotaExceeded):
            self._upload_test_document()

        self.assertEqual(
            Document.objects.count(), test_document_count + 1
        )
        self.assertEqual(
            DocumentFile.objects.count(), test_document_file_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        _test_document = Document.objects.first()
        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, _test_document)
        self.assertEqual(events[0].target, _test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

    def test_user_test_document_type_test(self):
        self._test_quota = DocumentSizeQuota.create(
            document_size_limit=0.01,
            document_type_all=False,
            document_type_ids=(self._test_document_type.pk,),
            group_ids=(),
            user_all=False,
            user_ids=(self._test_case_user.pk,)
        )

        test_document_count = Document.objects.count()
        test_document_file_count = DocumentFile.objects.count()

        self._clear_events()

        with self.assertRaises(expected_exception=QuotaExceeded):
            self._upload_test_document(user=self._test_case_user)

        self.assertEqual(
            Document.objects.count(), test_document_count + 1
        )
        self.assertEqual(
            DocumentFile.objects.count(), test_document_file_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        _test_document = Document.objects.first()
        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, _test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

    def test_group_test_document_type_test(self):
        self._create_test_group()
        self._test_case_user.groups.add(self._test_group)

        self._test_quota = DocumentSizeQuota.create(
            document_size_limit=0.01,
            document_type_all=False,
            document_type_ids=(self._test_document_type.pk,),
            group_ids=(self._test_group.pk,),
            user_all=False,
            user_ids=()
        )

        test_document_count = Document.objects.count()
        test_document_file_count = DocumentFile.objects.count()

        self._clear_events()

        with self.assertRaises(expected_exception=QuotaExceeded):
            self._upload_test_document(user=self._test_case_user)

        self.assertEqual(
            Document.objects.count(), test_document_count + 1
        )
        self.assertEqual(
            DocumentFile.objects.count(), test_document_file_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        _test_document = Document.objects.first()
        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, _test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

    def test_allow(self):
        self._test_quota = DocumentSizeQuota.create(
            document_size_limit=0.01,
            document_type_all=False,
            document_type_ids=(),
            group_ids=(),
            user_all=False,
            user_ids=()
        )

        self._clear_events()

        self._upload_test_document(user=self._test_case_user)

        events = self._get_test_events()
        self.assertEqual(events.count(), 6)

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, self._test_document)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, self._test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, self._test_document)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, self._test_document_file)
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        self.assertEqual(events[3].action_object, self._test_document)
        self.assertEqual(events[3].actor, self._test_case_user)
        self.assertEqual(events[3].target, self._test_document_version)
        self.assertEqual(events[3].verb, event_document_version_created.id)

        self.assertEqual(
            events[4].action_object, self._test_document_version
        )
        self.assertEqual(events[4].actor, self._test_case_user)
        self.assertEqual(events[4].target, self._test_document_version_page)
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[5].action_object, self._test_document)
        self.assertEqual(events[5].actor, self._test_case_user)
        self.assertEqual(events[5].target, self._test_document_version)
        self.assertEqual(events[5].verb, event_document_version_edited.id)

    def test_super_user_restriction(self):
        self._create_test_super_user()

        self._test_quota = DocumentSizeQuota.create(
            document_size_limit=0.01,
            document_type_all=True,
            document_type_ids=(),
            group_ids=(),
            user_all=True,
            user_ids=()
        )

        self._clear_events()

        self._upload_test_document(user=self._test_super_user)

        events = self._get_test_events()
        self.assertEqual(events.count(), 6)

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_super_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, self._test_document)
        self.assertEqual(events[1].actor, self._test_super_user)
        self.assertEqual(events[1].target, self._test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, self._test_document)
        self.assertEqual(events[2].actor, self._test_super_user)
        self.assertEqual(events[2].target, self._test_document_file)
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        self.assertEqual(events[3].action_object, self._test_document)
        self.assertEqual(events[3].actor, self._test_super_user)
        self.assertEqual(events[3].target, self._test_document_version)
        self.assertEqual(events[3].verb, event_document_version_created.id)

        self.assertEqual(
            events[4].action_object, self._test_document_version
        )
        self.assertEqual(events[4].actor, self._test_super_user)
        self.assertEqual(events[4].target, self._test_document_version_page)
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[5].action_object, self._test_document)
        self.assertEqual(events[5].actor, self._test_super_user)
        self.assertEqual(events[5].target, self._test_document_version)
        self.assertEqual(events[5].verb, event_document_version_edited.id)
